from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from .._utils import SIZE_FACTOR, to_rgba
from ..doctools import document
from ..exceptions import PlotnineError
from .geom import geom
from .geom_point import geom_point
from .geom_polygon import geom_polygon

if typing.TYPE_CHECKING:
    from typing import Any

    import numpy.typing as npt
    from matplotlib.axes import Axes
    from matplotlib.offsetbox import DrawingArea
    from matplotlib.patches import PathPatch
    from shapely.geometry.polygon import LinearRing, Polygon

    from plotnine import aes
    from plotnine.coords.coord import coord
    from plotnine.iapi import panel_view
    from plotnine.layer import layer
    from plotnine.typing import DataLike


@document
class geom_map(geom):
    """
    Draw map feature

    The map feature are drawn without any special projections.

    {usage}

    Parameters
    ----------
    {common_parameters}

    Notes
    -----
    This geom is best suited for plotting a shapefile read into
    geopandas dataframe. The dataframe should have a `geometry`
    column.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "#111111",
        "fill": "#333333",
        "linetype": "solid",
        "shape": "o",
        "size": 0.5,
        "stroke": 0.5,
    }
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
    }
    REQUIRED_AES = {"geometry"}

    def __init__(
        self,
        mapping: aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        geom.__init__(self, mapping, data, **kwargs)
        # Almost all geodataframes loaded from shapefiles
        # have a geometry column.
        if "geometry" not in self.mapping:
            self.mapping["geometry"] = "geometry"

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if not len(data):
            return data

        # Remove any NULL geometries, and remember
        # All the non-Null shapes in a shapefile are required to be
        # of the same shape type.
        bool_idx = np.array([g is not None for g in data["geometry"]])
        if not np.all(bool_idx):
            data = data.loc[bool_idx]

        # Add polygon limits. Scale training uses them
        try:
            bounds = data["geometry"].bounds
        except AttributeError:
            # The geometry is not a GeoSeries
            # Bounds calculation is extracted from
            # geopandas.base.GeoPandasBase.bounds
            bounds = pd.DataFrame(
                np.array([x.bounds for x in data["geometry"]]),
                columns=["xmin", "ymin", "xmax", "ymax"],
                index=data.index,
            )
        else:
            bounds.rename(
                columns={
                    "minx": "xmin",
                    "maxx": "xmax",
                    "miny": "ymin",
                    "maxy": "ymax",
                },
                inplace=True,
            )

        data = pd.concat([data, bounds], axis=1)
        return data

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: coord,
        ax: Axes,
        **params: Any,
    ):
        if not len(data):
            return

        data.loc[data["color"].isna(), "color"] = "none"
        data.loc[data["fill"].isna(), "fill"] = "none"
        data["fill"] = to_rgba(data["fill"], data["alpha"])

        geom_type = data.geometry.iloc[0].geom_type
        if geom_type in ("Polygon", "MultiPolygon"):
            from matplotlib.collections import PatchCollection

            data["size"] *= SIZE_FACTOR
            patches = [PolygonPatch(g) for g in data["geometry"]]
            coll = PatchCollection(
                patches,
                edgecolor=data["color"],
                facecolor=data["fill"],
                linestyle=data["linetype"],
                linewidth=data["size"],
                zorder=params["zorder"],
                rasterized=params["raster"],
            )
            ax.add_collection(coll)
        elif geom_type == "Point":
            # Extract point coordinates from shapely geom
            # and plot with geom_point
            arr = np.array([list(g.coords)[0] for g in data["geometry"]])
            data["x"] = arr[:, 0]
            data["y"] = arr[:, 1]
            for _, gdata in data.groupby("group"):
                gdata.reset_index(inplace=True, drop=True)
                gdata.is_copy = None
                geom_point.draw_group(gdata, panel_params, coord, ax, **params)
        elif geom_type == "MultiPoint":
            # Where n is the length of the dataframe (no. of multipoints),
            #       m is the number of all points in all multipoints
            #
            # - MultiPoint -> List of Points (tuples) (n -> m)
            # - Explode the list, to create a dataframe were each point
            #      is associated with the right aesthetics (n -> m)
            # - Create x & y columns from the points (m -> m)
            data["points"] = [
                [p.coords[0] for p in mp.geoms] for mp in data["geometry"]
            ]
            data = data.explode("points", ignore_index=True)
            data["x"] = [p[0] for p in data["points"]]
            data["y"] = [p[1] for p in data["points"]]
            geom_point.draw_group(data, panel_params, coord, ax, **params)
        elif geom_type in ("LineString", "MultiLineString"):
            from matplotlib.collections import LineCollection

            data["size"] *= SIZE_FACTOR
            data["color"] = to_rgba(data["color"], data["alpha"])
            segments = []
            for g in data["geometry"]:
                if g.geom_type == "LineString":
                    segments.append(g.coords)
                else:
                    segments.extend(_g.coords for _g in g.geoms)

            coll = LineCollection(
                segments,
                edgecolor=data["color"],
                linewidth=data["size"],
                linestyle=data["linetype"],
                zorder=params["zorder"],
                rasterized=params["raster"],
            )
            ax.add_collection(coll)
        else:
            raise TypeError(f"Could not plot geometry of type '{geom_type}'")

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: layer
    ) -> DrawingArea:
        """
        Draw a rectangle in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        data["size"] = data["stroke"]
        del data["stroke"]
        return geom_polygon.draw_legend(data, da, lyr)


def PolygonPatch(
    obj: Polygon,
) -> PathPatch:
    """
    Return a Matplotlib patch from a Polygon/MultiPolygon Geometry

    Parameters
    ----------
    obj : shapley.geometry.Polygon | shapley.geometry.MultiPolygon
        A Polygon or MultiPolygon to create a patch for description

    Returns
    -------
    result : matplotlib.patches.PathPatch
        A patch representing the shapely geometry

    Notes
    -----
    This functionality was originally provided by the descartes package
    by Sean Gillies (BSD license, https://pypi.org/project/descartes)
    which is nolonger being maintained.
    """
    from matplotlib.patches import PathPatch
    from matplotlib.path import Path

    def cw_coords(ring: LinearRing) -> npt.NDArray[Any]:
        """
        Return Clockwise array coordinates

        Parameters
        ----------
        ring: shapely.geometry.polygon.LinearRing
            LinearRing

        Returns
        -------
        out: ndarray
            (n x 2) array of coordinate points.
        """
        if ring.is_ccw:
            return np.asarray(ring.coords)[:, :2][::-1]
        return np.asarray(ring.coords)[:, :2]

    def ccw_coords(ring: LinearRing) -> npt.NDArray[Any]:
        """
        Return Counter Clockwise array coordinates

        Parameters
        ----------
        ring: shapely.geometry.polygon.LinearRing
            LinearRing

        Returns
        -------
        out: ndarray
            (n x 2) array of coordinate points.
        """
        if ring.is_ccw:
            return np.asarray(ring.coords)[:, :2]
        return np.asarray(ring.coords)[:, :2][::-1]

    # The interiors are holes in the Polygon
    # MPL draws a hole if the vertex points are specified
    # in an opposite direction. So we use Clockwise for
    # the exterior/shell and Counter-Clockwise for any
    # interiors/holes
    if obj.geom_type == "Polygon":
        _exterior = [Path(cw_coords(obj.exterior))]
        _interior = [Path(ccw_coords(ring)) for ring in obj.interiors]
    else:
        # A MultiPolygon has one or more Polygon geoms.
        # Concatenate the exterior of all the Polygons
        # and the interiors
        _exterior = []
        _interior = []
        for p in obj.geoms:
            _exterior.append(Path(cw_coords(p.exterior)))
            _interior.extend([Path(ccw_coords(ring)) for ring in p.interiors])

    path = Path.make_compound_path(*_exterior, *_interior)
    return PathPatch(path)


def check_geopandas():
    try:
        import geopandas  # noqa: F401
    except ImportError as e:
        msg = "geom_map requires geopandas. Please install geopandas."
        raise PlotnineError(msg) from e



from __future__ import annotations

import typing
from warnings import warn

from .._utils.registry import alias
from ..doctools import document
from ..exceptions import PlotnineWarning
from .scale_discrete import scale_discrete

if typing.TYPE_CHECKING:
    from plotnine.typing import ScaleBreaksRaw


@document
class _scale_manual(scale_discrete):
    """
    Abstract class for manual scales

    Parameters
    ----------
    {superclass_parameters}
    """

    def __init__(self, values, **kwargs):
        # Match the values of the scale with the breaks (if given)
        if "breaks" in kwargs:
            from collections.abc import Sized

            breaks: ScaleBreaksRaw = kwargs["breaks"]
            if isinstance(breaks, Sized) and len(breaks) == len(values):
                values = dict(zip(breaks, values))

        self._values = values
        scale_discrete.__init__(self, **kwargs)

    def palette(self, value):
        max_n = len(self._values)
        if value > max_n:
            msg = (
                f"The palette of {self.__class__.__name__} can return a "
                f"maximum of {max_n} values. {value} were requested from it."
            )
            warn(msg, PlotnineWarning)
        return self._values


@document
class scale_color_manual(_scale_manual):
    """
    Custom discrete color scale

    Parameters
    ----------
    values : array_like | dict
        Colors that make up the palette. The values will be matched with
        the `limits` of the scale or the `breaks` if provided.
        If it is a dict then it should map data values to colors.

    {superclass_parameters}
    """

    _aesthetics = ["color"]
    na_value = "#7F7F7F"


@document
class scale_fill_manual(_scale_manual):
    """
    Custom discrete fill scale

    Parameters
    ----------
    values : array_like | dict
        Colors that make up the palette. The values will be matched with
        the `limits` of the scale or the `breaks` if provided.
        If it is a dict then it should map data values to colors.
    {superclass_parameters}
    """

    _aesthetics = ["fill"]
    na_value = "#7F7F7F"


@document
class scale_shape_manual(_scale_manual):
    """
    Custom discrete shape scale

    Parameters
    ----------
    values : array_like | dict
        Shapes that make up the palette. See
        [](`matplotlib.markers`) for list of all possible
        shapes. The values will be matched with the `limits`
        of the scale or the `breaks` if provided.
        If it is a dict then it should map data values to shapes.
    {superclass_parameters}

    See Also
    --------
    [](`matplotlib.markers`)
    """

    _aesthetics = ["shape"]


@document
class scale_linetype_manual(_scale_manual):
    """
    Custom discrete linetype scale

    Parameters
    ----------
    values : list | dict
        Linetypes that make up the palette.
        Possible values of the list are:

        1. Strings like

        ```python
        'solid'                # solid line
        'dashed'               # dashed line
        'dashdot'              # dash-dotted line
        'dotted'               # dotted line
        'None' or ' ' or ''    # draw nothing
        ```

        2. Tuples of the form (offset, (on, off, on, off, ....))
           e.g. (0, (1, 1)), (1, (2, 2)), (2, (5, 3, 1, 3))

        The values will be matched with the `limits` of the scale
        or the `breaks` if provided.
        If it is a dict then it should map data values to linetypes.
    {superclass_parameters}

    See Also
    --------
    [](`matplotlib.markers`)
    """

    _aesthetics = ["linetype"]

    def map(self, x, limits=None):
        result = super().map(x, limits)
        # Ensure that custom linetypes are tuples, so that they can
        # be properly inserted and extracted from the dataframe
        if len(result) and hasattr(result[0], "__hash__"):
            result = [x if isinstance(x, str) else tuple(x) for x in result]
        return result


@document
class scale_alpha_manual(_scale_manual):
    """
    Custom discrete alpha scale

    Parameters
    ----------
    values : array_like | dict
        Alpha values (in the [0, 1] range) that make up
        the palette. The values will be matched with the
        `limits` of the scale or the `breaks` if provided.
        If it is a dict then it should map data values to alpha
        values.
    {superclass_parameters}
    """

    _aesthetics = ["alpha"]


@document
class scale_size_manual(_scale_manual):
    """
    Custom discrete size scale

    Parameters
    ----------
    values : array_like | dict
        Sizes that make up the palette. The values will be matched
        with the `limits` of the scale or the `breaks` if provided.
        If it is a dict then it should map data values to sizes.
    {superclass_parameters}
    """

    _aesthetics = ["size"]


# American to British spelling
@alias
class scale_colour_manual(scale_color_manual):
    pass



from __future__ import annotations

from matplotlib.offsetbox import (
    AnchoredOffsetbox,
    AuxTransformBox,
    DrawingArea,
)
from matplotlib.patches import bbox_artist as mbbox_artist
from matplotlib.transforms import Affine2D, Bbox

from .patches import InsideStrokedRectangle

DEBUG = False


# for debugging use
def _bbox_artist(*args, **kwargs):
    if DEBUG:
        mbbox_artist(*args, **kwargs)


class ColoredDrawingArea(DrawingArea):
    """
    A Drawing Area with a background color
    """

    def __init__(
        self,
        width: float,
        height: float,
        xdescent=0.0,
        ydescent=0.0,
        clip=True,
        color="none",
    ):
        super().__init__(width, height, xdescent, ydescent, clip=clip)

        self.patch = InsideStrokedRectangle(
            (0, 0),
            width=width,
            height=height,
            facecolor=color,
            edgecolor="none",
            linewidth=0,
            antialiased=False,
        )
        self.add_artist(self.patch)


# Fix AuxTransformBox, Adds a dpi_transform
# See https://github.com/matplotlib/matplotlib/pull/7344
class DPICorAuxTransformBox(AuxTransformBox):
    """
    DPI Corrected AuxTransformBox
    """

    def __init__(self, aux_transform):
        super().__init__(aux_transform)
        self.dpi_transform = Affine2D()
        self._dpi_corrected = False

    def get_transform(self):
        """
        Return the [](`~matplotlib.transforms.Transform`) applied
        to the children
        """
        return (
            self.aux_transform
            + self.dpi_transform
            + self.ref_offset_transform
            + self.offset_transform
        )

    def _correct_dpi(self, renderer):
        if not self._dpi_corrected:
            dpi_cor = renderer.points_to_pixels(1.0)
            self.dpi_transform.clear()
            self.dpi_transform.scale(dpi_cor, dpi_cor)
            self._dpi_corrected = True

    def get_bbox(self, renderer):
        self._correct_dpi(renderer)
        return super().get_bbox(renderer)

    def draw(self, renderer):
        """
        Draw the children
        """
        self._correct_dpi(renderer)
        for c in self.get_children():
            c.draw(renderer)

        _bbox_artist(self, renderer, fill=False, props=dict(pad=0.0))
        self.stale = False


class FlexibleAnchoredOffsetbox(AnchoredOffsetbox):
    """
    An AnchoredOffsetbox that accepts x, y location
    """

    def __init__(self, xy_loc: tuple[float, float] = (0.5, 0.5), **kwargs):
        if "loc" in kwargs:
            raise ValueError(
                "FlexibleAnchoredOffsetbox does not use the 'loc' parameter"
            )

        super().__init__(loc="center", **kwargs)
        self.xy_loc = xy_loc

    def get_offset(self, bbox, renderer):  # type: ignore
        pad = self.borderpad * renderer.points_to_pixels(
            self.prop.get_size_in_points()
        )
        parentbbox = self.get_bbox_to_anchor()
        _bbox = Bbox.from_bounds(0, 0, bbox.width, bbox.height)
        container = parentbbox.padded(-pad)
        x0, y0 = _bbox.anchored(self.xy_loc, container=container).p0
        return x0 - bbox.x0, y0 - bbox.y0



from .elements import element_blank
from .theme import theme
from .theme_bw import theme_bw


class theme_tufte(theme_bw):
    """
    Tufte Maximal Data, Minimal Ink Theme

    Theme based on Chapter 6 Data-Ink Maximization and Graphical
    Design of Edward Tufte *The Visual Display of Quantitative
    Information*. No border, no axis lines, no grids. This theme
    works best in combination with
    [](`~plotnine.geoms.geom_rug.geom_rug`).

    The default font family is set to "serif" as he uses serif
    fonts for labels in _The Visual Display of Quantitative
    Information_. The serif font used by Tufte in his books is
    a variant of Bembo, while the sans serif font is Gill Sans.
    If these fonts are installed on your system, consider setting
    them explicitly via the argument `base_family`.

    Parameters
    ----------
    base_size : int
        Base font size. All text sizes are a scaled versions of
        the base font size.
    base_family : str
        Base font family. If `None`, use [](`plotnine.options.base_family`).
    ticks: bool
        Whether to show axis ticks.

    References
    ----------
    Tufte, Edward R. (2001) The Visual Display of Quantitative
    Information, Chapter 6.

    Translated from the R ggthemes package by hyiltiz <hyiltiz@gmail.com>.
    Released under GNU GPL v2 license or later.
    """

    def __init__(self, base_size=11, base_family=None, ticks=True):
        theme_bw.__init__(self, base_size, base_family)
        d = {} if ticks else {"axis_ticks": element_blank()}
        self += theme(
            axis_line=element_blank(),
            legend_background=element_blank(),
            legend_key=element_blank(),
            panel_background=element_blank(),
            panel_border=element_blank(),
            panel_grid=element_blank(),
            plot_background=element_blank(),
            strip_background=element_blank(),
            **d,  # pyright: ignore
        )



from __future__ import annotations

import typing

from matplotlib.transforms import Affine2D, Bbox

from .transforms import ZEROS_BBOX

if typing.TYPE_CHECKING:
    from matplotlib.artist import Artist
    from matplotlib.axes import Axes
    from matplotlib.backend_bases import RendererBase
    from matplotlib.figure import Figure
    from matplotlib.transforms import Transform


def bbox_in_figure_space(
    artist: Artist, fig: Figure, renderer: RendererBase
) -> Bbox:
    """
    Bounding box of artist in figure coordinates
    """
    box = artist.get_window_extent(renderer) or ZEROS_BBOX
    return fig.transFigure.inverted().transform_bbox(box)


def tight_bbox_in_figure_space(
    artist: Artist, fig: Figure, renderer: RendererBase
) -> Bbox:
    """
    Bounding box of artist and its children in figure coordinates
    """
    box = artist.get_tightbbox(renderer) or ZEROS_BBOX
    return fig.transFigure.inverted().transform_bbox(box)


def bbox_in_axes_space(
    artist: Artist, ax: Axes, renderer: RendererBase
) -> Bbox:
    """
    Bounding box of artist in figure coordinates
    """
    box = artist.get_window_extent(renderer) or ZEROS_BBOX
    return ax.transAxes.inverted().transform_bbox(box)


def pts_in_figure_space(fig: Figure, pts: float) -> float:
    """
    Points in figure coordinates
    """
    return fig.transFigure.inverted().transform([0, pts])[1]


def get_transPanels(fig: Figure) -> Transform:
    """
    Coordinate system of the Panels (facets) area

    (0, 0) is the bottom-left of the bottom-left panel and
    (1, 1) is the top right of the top-right panel.

    The subplot parameters must be set before calling this function.
    i.e. fig.subplots_adjust should have been called.
    """
    # Contains the layout information from which the panel area
    # is derived
    params = fig.subplotpars

    # Figure width & height in display coordinates
    W, H = fig.bbox.width, fig.bbox.height

    # 1. The panels occupy space that is smaller than the figure
    # 2. That space is contained within the figure
    # We create a transform that represent these separable aspects
    # (but order matters), and use to transform transFigure
    sx, sy = params.right - params.left, params.top - params.bottom
    dx, dy = params.left * W, params.bottom * H
    transFiguretoPanels = Affine2D().scale(sx, sy).translate(dx, dy)
    return fig.transFigure + transFiguretoPanels



from __future__ import annotations

from contextlib import suppress

from ..doctools import document
from .scale_continuous import scale_continuous


@document
class scale_datetime(scale_continuous):
    """
    Base class for all date/datetime scales

    Parameters
    ----------
    date_breaks : str, default=None
        A string giving the distance between major breaks.
        For example `'2 weeks'`, `'5 years'`. If specified,
        `date_breaks` takes precedence over
        `breaks`.
    date_labels : str, default=None
        Format string for the labels.
        See [strftime](:ref:`strftime-strptime-behavior`).
        If specified, `date_labels` takes precedence over
        `labels`.
    date_minor_breaks : str, default=None
        A string giving the distance between minor breaks.
        For example `'2 weeks'`, `'5 years'`. If specified,
        `date_minor_breaks` takes precedence over
        `minor_breaks`.
    {superclass_parameters}
    """

    _trans = "datetime"

    def __init__(self, **kwargs):
        from mizani.breaks import breaks_date
        from mizani.labels import label_date

        # Permit the use of the general parameters for
        # specifying the format strings
        with suppress(KeyError):
            breaks = kwargs["breaks"]
            if isinstance(breaks, str):
                kwargs["breaks"] = breaks_date(width=breaks)

        with suppress(KeyError):
            minor_breaks = kwargs["minor_breaks"]
            if isinstance(minor_breaks, str):
                kwargs["minor_breaks"] = breaks_date(width=minor_breaks)

        # Using the more specific parameters take precedence
        with suppress(KeyError):
            breaks_fmt = kwargs.pop("date_breaks")
            kwargs["breaks"] = breaks_date(breaks_fmt)

        with suppress(KeyError):
            labels_fmt = kwargs.pop("date_labels")
            kwargs["labels"] = label_date(labels_fmt)

        with suppress(KeyError):
            minor_breaks_fmt = kwargs.pop("date_minor_breaks")
            kwargs["minor_breaks"] = breaks_date(minor_breaks_fmt)

        scale_continuous.__init__(self, **kwargs)



"""
Routines to adjust subplot params so that subplots are
nicely fit in the figure. In doing so, only axis labels, tick labels, axes
titles and offsetboxes that are anchored to axes are currently considered.

Internally, this module assumes that the margins (left margin, etc.) which are
differences between `Axes.get_tightbbox` and `Axes.bbox` are independent of
Axes position. This may fail if `Axes.adjustable` is `datalim` as well as
such cases as when left or right margin are affected by xlabel.
"""

from __future__ import annotations

import typing
from copy import deepcopy
from dataclasses import dataclass

from ._plot_side_space import LRTBSpaces, WHSpaceParts, calculate_panel_spacing
from .utils import bbox_in_figure_space, get_transPanels

if typing.TYPE_CHECKING:
    from typing import Literal, TypeAlias

    from matplotlib.figure import Figure
    from matplotlib.text import Text
    from matplotlib.transforms import Transform

    from plotnine._mpl.offsetbox import FlexibleAnchoredOffsetbox
    from plotnine.facets.facet import facet
    from plotnine.iapi import legend_artists
    from plotnine.typing import TupleFloat2

    from .layout_engine import LayoutPack

    AxesLocation: TypeAlias = Literal[
        "all", "first_row", "last_row", "first_col", "last_col"
    ]


@dataclass
class GridSpecParams:
    """
    Gridspec Parameters
    """

    left: float
    right: float
    top: float
    bottom: float
    wspace: float
    hspace: float


@dataclass
class TightParams:
    """
    All parameters computed for the plotnine tight layout engine
    """

    sides: LRTBSpaces
    gullies: WHSpaceParts

    def __post_init__(self):
        self.params = GridSpecParams(
            left=self.sides.left,
            right=self.sides.right,
            top=self.sides.top,
            bottom=self.sides.bottom,
            wspace=self.gullies.wspace,
            hspace=self.gullies.hspace,
        )

    def to_aspect_ratio(
        self, facet: facet, ratio: float, parts: WHSpaceParts
    ) -> TightParams:
        """
        Modify TightParams to get a given aspect ratio
        """
        current_ratio = (parts.h * parts.H) / (parts.w * parts.W)
        increase_aspect_ratio = ratio > current_ratio
        if increase_aspect_ratio:  # Taller panel
            return self._reduce_width(facet, ratio, parts)
        else:
            return self._reduce_height(facet, ratio, parts)

    def _reduce_height(
        self, facet: facet, ratio: float, parts: WHSpaceParts
    ) -> TightParams:
        """
        Reduce the height of axes to get the aspect ratio
        """
        self = deepcopy(self)

        # New height w.r.t figure height
        h1 = ratio * parts.w * (parts.W / parts.H)

        # Half of the total vertical reduction w.r.t figure height
        dh = (parts.h - h1) * facet.nrow / 2

        # Reduce plot area height
        self.params.top -= dh
        self.params.bottom += dh
        self.params.hspace = parts.sh / h1

        # Add more vertical plot margin
        self.sides.t.plot_margin += dh
        self.sides.b.plot_margin += dh
        return self

    def _reduce_width(
        self, facet: facet, ratio: float, parts: WHSpaceParts
    ) -> TightParams:
        """
        Reduce the width of axes to get the aspect ratio
        """
        self = deepcopy(self)

        # New width w.r.t figure width
        w1 = (parts.h * parts.H) / (ratio * parts.W)

        # Half of the total horizontal reduction w.r.t figure width
        dw = (parts.w - w1) * facet.ncol / 2

        # Reduce width
        self.params.left += dw
        self.params.right -= dw
        self.params.wspace = parts.sw / w1

        # Add more horizontal margin
        self.sides.l.plot_margin += dw
        self.sides.r.plot_margin += dw
        return self


def get_plotnine_tight_layout(pack: LayoutPack) -> TightParams:
    """
    Compute tight layout parameters
    """
    sides = LRTBSpaces(pack)
    gullies = calculate_panel_spacing(pack, sides)
    tight_params = TightParams(sides, gullies)
    ratio = pack.facet._aspect_ratio()
    if ratio is not None:
        tight_params = tight_params.to_aspect_ratio(pack.facet, ratio, gullies)
    return tight_params


def set_figure_artist_positions(
    pack: LayoutPack,
    tparams: TightParams,
):
    """
    Set the x,y position of the artists around the panels
    """
    theme = pack.theme
    sides = tparams.sides
    params = tparams.params

    if pack.plot_title:
        ha = theme.getp(("plot_title", "ha"))
        pack.plot_title.set_y(sides.t.edge("plot_title"))
        horizonally_align_text_with_panels(pack.plot_title, params, ha, pack)

    if pack.plot_subtitle:
        ha = theme.getp(("plot_subtitle", "ha"))
        pack.plot_subtitle.set_y(sides.t.edge("plot_subtitle"))
        horizonally_align_text_with_panels(
            pack.plot_subtitle, params, ha, pack
        )

    if pack.plot_caption:
        ha = theme.getp(("plot_caption", "ha"), "right")
        pack.plot_caption.set_y(sides.b.edge("plot_caption"))
        horizonally_align_text_with_panels(pack.plot_caption, params, ha, pack)

    if pack.axis_title_x:
        ha = theme.getp(("axis_title_x", "ha"), "center")
        pack.axis_title_x.set_y(sides.b.edge("axis_title_x"))
        horizonally_align_text_with_panels(pack.axis_title_x, params, ha, pack)

    if pack.axis_title_y:
        va = theme.getp(("axis_title_y", "va"), "center")
        pack.axis_title_y.set_x(sides.l.edge("axis_title_y"))
        vertically_align_text_with_panels(pack.axis_title_y, params, va, pack)

    if pack.legends:
        set_legends_position(pack.legends, tparams, pack.figure)


def horizonally_align_text_with_panels(
    text: Text, params: GridSpecParams, ha: str | float, pack: LayoutPack
):
    """
    Horizontal justification

    Reinterpret horizontal alignment to be justification about the panels.
    """
    if isinstance(ha, str):
        lookup = {
            "left": 0.0,
            "center": 0.5,
            "right": 1.0,
        }
        f = lookup[ha]
    else:
        f = ha

    box = bbox_in_figure_space(text, pack.figure, pack.renderer)
    x = params.left * (1 - f) + (params.right - box.width) * f
    text.set_x(x)
    text.set_horizontalalignment("left")


def vertically_align_text_with_panels(
    text: Text, params: GridSpecParams, va: str | float, pack: LayoutPack
):
    """
    Vertical justification

    Reinterpret vertical alignment to be justification about the panels.
    """
    if isinstance(va, str):
        lookup = {
            "top": 1.0,
            "center": 0.5,
            "baseline": 0.5,
            "center_baseline": 0.5,
            "bottom": 0.0,
        }
        f = lookup[va]
    else:
        f = va

    box = bbox_in_figure_space(text, pack.figure, pack.renderer)
    y = params.bottom * (1 - f) + (params.top - box.height) * f
    text.set_y(y)
    text.set_verticalalignment("bottom")


def set_legends_position(
    legends: legend_artists,
    tparams: TightParams,
    fig: Figure,
):
    """
    Place legend on the figure and justify is a required
    """

    def set_position(
        aob: FlexibleAnchoredOffsetbox,
        anchor_point: TupleFloat2,
        xy_loc: TupleFloat2,
        transform: Transform = fig.transFigure,
    ):
        """
        Place box (by the anchor point) at given xy location

        Parameters
        ----------
        aob :
           Offsetbox to place
        anchor_point :
            Point on the Offsefbox.
        xy_loc :
            Point where to place the offsetbox.
        transform :
            Transformation
        """
        aob.xy_loc = xy_loc
        aob.set_bbox_to_anchor(anchor_point, transform)  # type: ignore

    sides = tparams.sides
    params = fig.subplotpars

    if legends.right:
        j = legends.right.justification
        y = params.bottom * (1 - j) + (params.top - sides.r._legend_height) * j
        x = sides.r.edge("legend")
        set_position(legends.right.box, (x, y), (1, 0))

    if legends.left:
        j = legends.left.justification
        y = params.bottom * (1 - j) + (params.top - sides.l._legend_height) * j
        x = sides.l.edge("legend")
        set_position(legends.left.box, (x, y), (0, 0))

    if legends.top:
        j = legends.top.justification
        x = params.left * (1 - j) + (params.right - sides.t._legend_width) * j
        y = sides.t.edge("legend")
        set_position(legends.top.box, (x, y), (0, 1))

    if legends.bottom:
        j = legends.bottom.justification
        x = params.left * (1 - j) + (params.right - sides.b._legend_width) * j
        y = sides.b.edge("legend")
        set_position(legends.bottom.box, (x, y), (0, 0))

    # Inside legends are placed using the panels coordinate system
    if legends.inside:
        transPanels = get_transPanels(fig)
        for l in legends.inside:
            set_position(l.box, l.position, l.justification, transPanels)



from __future__ import annotations

import itertools
import os
import re
import typing
from functools import lru_cache
from textwrap import dedent, indent

if typing.TYPE_CHECKING:
    from typing import Any, Type, TypeVar

    from plotnine.geoms.geom import geom
    from plotnine.scales.scale import scale
    from plotnine.stats.stat import stat

    T = TypeVar("T")

# Markup that is robust for documentation needs grid tables
# using tabulate and it is only required when
# building documentation.
try:
    from tabulate import tabulate as table_function
except ImportError:
    from ._utils import simple_table as table_function


# Parameter arguments that are listed first in the geom and
# stat class signatures

common_geom_params = [
    "mapping",
    "data",
    "stat",
    "position",
    "na_rm",
    "inherit_aes",
    "show_legend",
    "raster",
]
common_geom_param_values = {
    "mapping": None,
    "data": None,
    "inherit_aes": True,
    "show_legend": None,
    "raster": False,
}

common_stat_params = ["mapping", "data", "geom", "position", "na_rm"]
common_stat_param_values = common_geom_param_values

# Templates for docstrings

GEOM_SIGNATURE_TPL = """
**Usage**

{signature}

""".strip()

AESTHETICS_TABLE_TPL = """
{table}

The **bold** aesthetics are required."""

STAT_SIGNATURE_TPL = """
**Usage**

{signature}

""".strip()

common_params_doc = {
    "mapping": """\
Aesthetic mappings created with [aes](:class:`plotnine.mapping.aes`). If \
specified and `inherit_aes=True`{.py}, it is combined with the default \
mapping for the plot. You must supply mapping if there is no plot mapping.""",
    "data": """\
The data to be displayed in this layer. If `None`{.py}, the data from \
from the `ggplot()`{.py} call is used. If specified, it overrides the \
data from the `ggplot()`{.py} call.""",
    "stat": """\
The statistical transformation to use on the data for this layer. \
If it is a string, it must be the registered and known to Plotnine.""",
    "position": """\
Position adjustment. If it is a string, it must be registered and \
known to Plotnine.""",
    "na_rm": """\
If `False`{.py}, removes missing values with a warning. If `True`{.py} \
silently removes missing values.""",
    "inherit_aes": """\
If `False`{.py}, overrides the default aesthetics.""",
    "show_legend": """\
Whether this layer should be included in the legends. `None`{.py} the \
default, includes any aesthetics that are mapped. If a [](:class:`bool`), \
`False`{.py} never includes and `True`{.py} always includes. A \
[](:class:`dict`) can be used to *exclude* specific aesthetis of the layer \
from showing in the legend. e.g `show_legend={'color': False}`{.py}, \
any other aesthetic are included by default.""",
    "raster": """\
If `True`, draw onto this layer a raster (bitmap) object even if\
the final image is in vector format.""",
}


GEOM_PARAMS_TPL = """
mapping : ~plotnine.mapping.aes, default=None
    {mapping}
    {_aesthetics_doc}
data : ~pandas.DataFrame, default=None
    {data}
stat : str | ~plotnine.stats.stat.stat, default="{default_stat}"
    {stat}
position : str | ~plotnine.positions.position.position, \
default="{default_position}"
    {position}
na_rm : bool, default={default_na_rm}
    {na_rm}
inherit_aes : bool, default={default_inherit_aes}
    {inherit_aes}
show_legend : bool | dict, default=None
    {show_legend}
raster : bool, default={default_raster}
    {raster}
"""

STAT_PARAMS_TPL = """
mapping : ~plotnine.mapping.aes, default=None
    {mapping}
    {_aesthetics_doc}
data : ~pandas.DataFrame, default=None
    {data}
geom : str | ~plotnine.geoms.geom.geom, default="{default_geom}"
    {stat}
position : str | ~plotnine.positions.position.position, \
default="{default_position}"
    {position}
na_rm : bool, default={default_na_rm}
    {na_rm}
"""

geom_kwargs = """\
**kwargs: Any
    Aesthetics or parameters used by the `stat`.
"""

stat_kwargs = """\
**kwargs: Any
    Aesthetics or parameters used by the `geom`.
"""

DOCSTRING_SECTIONS = {
    "parameters",
    "see also",
    "note",
    "notes",
    "example",
    "examples",
}

PARAM_PATTERN = re.compile(r"\s*" r"([_A-Za-z]\w*)" r"\s:\s")
SECTIONS_PATTERN = re.compile(
    r"\n(?P<section>(?:\w+|(\w+\s\w+)+))\s*"  # section name
    r"\n-{3,}\n",  # underline
)
GENERATING_QUARTODOC = os.environ.get("GENERATING_QUARTODOC")


def dict_to_table(header: tuple[str, str], contents: dict[str, str]) -> str:
    """
    Convert dict to an (n x 2) table

    Parameters
    ----------
    header : tuple
        Table header. Should have a length of 2.
    contents : dict
        The key becomes column 1 of table and the
        value becomes column 2 of table.

    Examples
    --------
    >>> d = {"alpha": 1, "color": "blue", "fill": None}
    >>> print(dict_to_table(("Aesthetic", "Default Value"), d))
    Aesthetic  Default Value
    ---------  -------------
    alpha      `1`
    color      `'blue'`
    fill       `None`
    """
    rows = [
        (name, value if value == "" else f"`{value!r}`" "{.py}")
        for name, value in contents.items()
    ]
    return table_function(rows, headers=header, tablefmt="grid")


def make_signature(
    name: str,
    params: dict[str, Any],
    common_params: list[str],
    common_param_values: dict[str, Any],
) -> str:
    """
    Create a signature for a geom or stat

    Gets the DEFAULT_PARAMS (params) and creates are comma
    separated list of the `name=value` pairs. The common_params
    come first in the list, and they get take their values from
    either the params-dict or the common_geom_param_values-dict.
    """
    params_lst = []
    _common_params_lookup = set(common_params)
    it = itertools.chain(
        common_params, (p for p in params if p not in _common_params_lookup)
    )

    # preferred params come first
    for i, key in enumerate(it):
        value = params.get(key, common_param_values.get(key))
        if isinstance(value, str):
            value = f'"{value}"'
        params_lst.append(f"{key}={value}")
        if i == 1:
            params_lst.append("*")

    params_lst.append("**kwargs")

    # Format to a maximum width of 78 chars
    # It fails when a parameter declarations is longer than 78
    opening = f"{name}("
    params_string = ", ".join(params_lst)
    closing = ")"
    pad = " " * 4
    if len(opening) + len(params_string) > 78:
        line_pad = f"\n{pad}"
        # One parameter per line
        if len(params_string) > 74:
            params_string = f",{line_pad}".join(params_lst)
        params_string = f"{line_pad}{params_string}"
        closing = f"\n{closing}"
    sig = f"{opening}{params_string}{closing}"
    return indent(sig, pad)


@lru_cache(maxsize=256)
def docstring_section_lines(docstring: str, section_name: str) -> str:
    """
    Return a section of a numpydoc string

    Parameters
    ----------
    docstring :
        Docstring
    section_name :
        Name of section to return

    Returns
    -------
    :
        Section minus the header
    """
    lines = []
    inside_section = False
    underline = "-" * len(section_name)
    expect_underline = False
    for line in docstring.splitlines():
        _line = line.strip().lower()

        if expect_underline:
            expect_underline = False
            if _line == underline:
                inside_section = True
                continue

        if _line == section_name:
            expect_underline = True
        elif _line in DOCSTRING_SECTIONS:
            # next section
            break
        elif inside_section:
            lines.append(line)
    return "\n".join(lines)


def append_to_section(s: str, docstring: str, section: str) -> str:
    """
    Append string s to a section in the docstring
    """
    idx = -1
    found = False
    for m in SECTIONS_PATTERN.finditer(docstring):
        if section == m.group("section"):
            found = True
        elif found:
            idx = m.start()
            break

    if found:
        if idx == -1:
            s = f"\n{s}"
        top, bottom = docstring[:idx], docstring[idx:]
        docstring = f"{top}{s}{bottom}"

    return docstring


def docstring_parameters_section(obj: Any) -> str:
    """
    Return the parameters section of a docstring
    """
    return docstring_section_lines(obj.__doc__, "parameters")


def param_spec(line: str) -> str | None:
    """
    Identify and return parameter

    Parameters
    ----------
    line : str
        A line in the parameter section.

    Returns
    -------
    name : str | None
        Name of the parameter if the line for the parameter
        type specification and None otherwise.

    Examples
    --------
    >>> param_spec('line : str')
    breaks
    >>> param_spec("    A line in the parameter section.")
    """
    m = PARAM_PATTERN.match(line)
    return m.group(1) if m else None


def parameters_str_to_dict(param_section: str) -> dict[str, str]:
    """
    Convert a param section to a dict

    Parameters
    ----------
    param_section : str
        Text in the parameter section

    Returns
    -------
    d : dict
        Dictionary of the parameters in the order that they
        are described in the parameters section. The dict
        is of the form `{param: all_parameter_text}`.
        You can reconstruct the `param_section` from the
        keys of the dictionary.

    See Also
    --------
    plotnine.doctools.parameters_dict_to_str
    """
    d = {}
    previous_param = ""
    param_desc: list[str] = []
    for line in param_section.split("\n"):
        param = param_spec(line)
        if param:
            if previous_param:
                d[previous_param] = "\n".join(param_desc)
            param_desc = [line]
            previous_param = param
        elif param_desc:
            param_desc.append(line)

    if previous_param:
        d[previous_param] = "\n".join(param_desc)

    return d


def parameters_dict_to_str(d: dict[str, str]) -> str:
    """
    Convert a dict of param section to a string

    Parameters
    ----------
    d : dict
        Parameters and their descriptions in a docstring

    Returns
    -------
    param_section : str
        Text in the parameter section

    See Also
    --------
    plotnine.doctools.parameters_str_to_dict
    """
    return "\n".join(d.values())


def default_class_name(s: str | type | object) -> str:
    """
    Return the qualified name of s

    Only if s does not start with the prefix

    Examples
    --------
    >>> qualified_name('stat_bin')
    'stat_bin'
    >>> qualified_name(stat_bin)
    'stat_bin'
    >>> qualified_name(stat_bin())
    'stat_bin'
    """
    if isinstance(s, str):
        return s
    elif isinstance(s, type):
        s = s.__name__
    else:
        s = s.__class__.__name__
    return s


def document_geom(geom: type[geom]) -> type[geom]:
    """
    Create a structured documentation for the geom

    It replaces `{usage}`, `{common_parameters}` and
    `{aesthetics}` with generated documentation.
    """
    docstring = dedent(geom.__doc__ or "")
    docstring = append_to_section(geom_kwargs, docstring, "Parameters")

    # usage
    signature = make_signature(
        geom.__name__,
        geom.DEFAULT_PARAMS,
        common_geom_params,
        common_geom_param_values,
    )
    usage = GEOM_SIGNATURE_TPL.format(signature=signature)

    # aesthetics
    contents = {f"**{ae}**": "" for ae in sorted(geom.REQUIRED_AES)}
    if geom.DEFAULT_AES:
        d = geom.DEFAULT_AES.copy()
        d["group"] = ""  # All geoms understand the group aesthetic
        contents.update(sorted(d.items()))

    table = dict_to_table(("Aesthetic", "Default value"), contents)
    aesthetics_table = AESTHETICS_TABLE_TPL.format(table=table)
    tpl = dedent(geom._aesthetics_doc).strip()
    aesthetics_doc = tpl.format(aesthetics_table=aesthetics_table)
    aesthetics_doc = indent(aesthetics_doc, " " * 4)

    # common_parameters
    d = geom.DEFAULT_PARAMS
    common_parameters = GEOM_PARAMS_TPL.format(
        default_stat=default_class_name(d["stat"]),
        default_position=default_class_name(d["position"]),
        default_na_rm=d["na_rm"],
        default_inherit_aes=d.get("inherit_aes", True),
        default_raster=d.get("raster", False),
        _aesthetics_doc=aesthetics_doc,
        **common_params_doc,
    ).strip()

    docstring = docstring.replace("{usage}", usage)
    docstring = docstring.replace("{common_parameters}", common_parameters)
    geom.__doc__ = docstring
    return geom


def document_stat(stat: type[stat]) -> type[stat]:
    """
    Create a structured documentation for the stat

    It replaces `{usage}`, `{common_parameters}` and
    `{aesthetics}` with generated documentation.
    """
    # Dedented so that it lineups (in sphinx) with the part
    # generated parts when put together
    docstring = dedent(stat.__doc__ or "")
    docstring = append_to_section(stat_kwargs, docstring, "Parameters")

    # usage:
    signature = make_signature(
        stat.__name__,
        stat.DEFAULT_PARAMS,
        common_stat_params,
        common_stat_param_values,
    )
    usage = STAT_SIGNATURE_TPL.format(signature=signature)

    # aesthetics
    contents = {f"**{ae}**": "" for ae in sorted(stat.REQUIRED_AES)}
    contents.update(sorted(stat.DEFAULT_AES.items()))
    table = dict_to_table(("Aesthetic", "Default value"), contents)
    aesthetics_table = AESTHETICS_TABLE_TPL.format(table=table)
    tpl = dedent(stat._aesthetics_doc).strip()
    aesthetics_doc = tpl.replace("{aesthetics_table}", aesthetics_table)
    aesthetics_doc = indent(aesthetics_doc, " " * 4)

    # common_parameters
    d = stat.DEFAULT_PARAMS
    common_parameters = STAT_PARAMS_TPL.format(
        default_geom=default_class_name(d["geom"]),
        default_position=default_class_name(d["position"]),
        default_na_rm=d["na_rm"],
        _aesthetics_doc=aesthetics_doc,
        **common_params_doc,
    ).strip()

    docstring = docstring.replace("{usage}", usage)
    docstring = docstring.replace("{common_parameters}", common_parameters)
    stat.__doc__ = docstring
    return stat


def document_scale(cls: type[scale]) -> type[scale]:
    """
    Create a documentation for a scale

    Import the superclass parameters

    It replaces `{superclass_parameters}` with the documentation
    of the parameters from the superclass.

    Parameters
    ----------
    cls : type
        A scale class

    Returns
    -------
    cls : type
        The scale class with a modified docstring.
    """
    params_list = []
    # Get set of cls params
    cls_param_string = docstring_parameters_section(cls)
    cls_param_dict = parameters_str_to_dict(cls_param_string)
    cls_params = set(cls_param_dict.keys())

    for i, base in enumerate(cls.__bases__):
        # Get set of base class params
        base_param_string = param_string = docstring_parameters_section(base)
        base_param_dict = parameters_str_to_dict(base_param_string)
        base_params = set(base_param_dict.keys())

        # Remove duplicate params from the base class
        duplicate_params = base_params & cls_params
        for param in duplicate_params:
            del base_param_dict[param]

        if duplicate_params:
            param_string = parameters_dict_to_str(base_param_dict)

        # Accumulate params of base case
        if i == 0:
            # Compensate for the indentation of the
            # {superclass_parameters} string
            param_string = param_string.strip()
        params_list.append(param_string)

        # Prevent the next base classes from bringing in the
        # same parameters.
        cls_params |= base_params

    # Fill in the processed superclass parameters
    superclass_parameters = "\n".join(params_list)
    cls_doc = cls.__doc__ or ""  # for typechecker
    cls.__doc__ = cls_doc.format(superclass_parameters=superclass_parameters)
    return cls


DOC_FUNCTIONS = {
    "geom": document_geom,
    "stat": document_stat,
    "scale": document_scale,
}


def document(cls: Type[T]) -> Type[T]:
    """
    Document a plotnine class

    To be used as a decorator
    """
    if cls.__doc__ is None:
        return cls

    baseclass_name = cls.mro()[-3].__name__

    try:
        return DOC_FUNCTIONS[baseclass_name](cls)
    except KeyError:
        return cls



from __future__ import annotations

import typing
from warnings import warn

import numpy as np
import pandas as pd

from .._utils import order_as_data_mapping
from ..doctools import document
from ..exceptions import PlotnineWarning
from ..mapping import aes
from .geom import geom
from .geom_path import geom_path
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    from typing import Any

    from matplotlib.axes import Axes

    from plotnine.coords.coord import coord
    from plotnine.iapi import panel_view
    from plotnine.typing import DataLike


@document
class geom_hline(geom):
    """
    Horizontal line

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_AES = {
        "color": "black",
        "linetype": "solid",
        "size": 0.5,
        "alpha": 1,
    }
    REQUIRED_AES = {"yintercept"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "inherit_aes": False,
    }

    draw_legend = staticmethod(geom_path.draw_legend)
    legend_key_size = staticmethod(geom_path.legend_key_size)

    def __init__(
        self,
        mapping: aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        data, mapping = order_as_data_mapping(data, mapping)
        yintercept = kwargs.pop("yintercept", None)
        if yintercept is not None:
            if mapping:
                warn(
                    "The 'yintercept' parameter has overridden "
                    "the aes() mapping.",
                    PlotnineWarning,
                )
            data = pd.DataFrame({"yintercept": np.repeat(yintercept, 1)})
            mapping = aes(yintercept="yintercept")
            kwargs["show_legend"] = False

        geom.__init__(self, mapping, data, **kwargs)

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups
        """
        ranges = coord.backtransform_range(panel_params)
        data["y"] = data["yintercept"]
        data["yend"] = data["yintercept"]
        data["x"] = ranges.x[0]
        data["xend"] = ranges.x[1]
        data = data.drop_duplicates()

        for _, gdata in data.groupby("group"):
            gdata.reset_index(inplace=True)
            geom_segment.draw_group(gdata, panel_params, coord, ax, **params)



from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from .._utils import add_margins, cross_join, join_keys, match, ninteraction
from ..exceptions import PlotnineError
from .facet import (
    add_missing_facets,
    combine_vars,
    eval_facet_vars,
    facet,
    layout_null,
)
from .strips import Strips, strip

if typing.TYPE_CHECKING:
    from typing import Literal, Optional, Sequence

    from matplotlib.axes import Axes

    from plotnine.iapi import layout_details
    from plotnine.typing import FacetSpaceRatios


class facet_grid(facet):
    """
    Wrap 1D Panels onto 2D surface

    Parameters
    ----------
    rows :
        Variable expressions along the rows of the facets/panels.
        Each expression is evaluated within the context of the dataframe.
    cols :
        Variable expressions along the columns of the facets/panels.
        Each expression is evaluated within the context of the dataframe.
    margins :
        variable names to compute margins for.
        True will compute all possible margins.
    space :
        Control the size of the  `x` or `y` sides of the panels.
        The size also depends to the `scales` parameter.

        If a string, it should be one of
        `['fixed', 'free', 'free_x', 'free_y']`{.py}.

        If a `dict`, it indicates the relative facet size ratios such as:

        ```python
        {"x": [1, 2], "y": [3, 1, 1]}
        ```

        This means that in the horizontal direction, the second panel
        will be twice the length of the first. In the vertical direction
        the top facet will be the 3 times longer then the second and
        third facets.

        Note that the number of dimensions in the list must equal the
        number of facets that will be produced.
    shrink :
        Whether to shrink the scales to the output of the
        statistics instead of the raw data.
    labeller :
        How to label the facets. A string value if it should be
        one of `["label_value", "label_both", "label_context"]`{.py}.
    as_table :
        If `True`, the facets are laid out like a table with
        the highest values at the bottom-right. If `False`
        the facets are laid out like a plot with the highest
        value a the top-right
    drop :
        If `True`, all factor levels not used in the data
        will automatically be dropped. If `False`, all
        factor levels will be shown, regardless of whether
        or not they appear in the data.
    """

    def __init__(
        self,
        rows: Optional[str | Sequence[str]] = None,
        cols: Optional[str | Sequence[str]] = None,
        *,
        margins: bool | list[str] = False,
        scales: Literal["fixed", "free", "free_x", "free_y"] = "fixed",
        space: (
            Literal["fixed", "free", "free_x", "free_y"] | FacetSpaceRatios
        ) = "fixed",
        shrink: bool = True,
        labeller: Literal[
            "label_value", "label_both", "label_context"
        ] = "label_value",
        as_table: bool = True,
        drop: bool = True,
    ):
        facet.__init__(
            self,
            scales=scales,
            shrink=shrink,
            labeller=labeller,
            as_table=as_table,
            drop=drop,
        )
        self.rows, self.cols = parse_grid_rows_cols(rows, cols)
        self.space = space
        self.margins = margins

    def _make_figure(self):
        import matplotlib.pyplot as plt
        from matplotlib.gridspec import GridSpec

        layout = self.layout
        space = self.space
        ratios = {}

        # Calculate the width (x) & height (y) ratios for space=free[xy]
        if isinstance(space, str):
            if space in {"free", "free_x"}:
                pidx: list[int] = (
                    layout.layout.sort_values("COL")
                    .drop_duplicates("COL")
                    .index.tolist()
                )
                panel_views = [layout.panel_params[i] for i in pidx]
                ratios["width_ratios"] = [
                    np.ptp(pv.x.range) for pv in panel_views
                ]

            if space in {"free", "free_y"}:
                pidx = (
                    layout.layout.sort_values("ROW")
                    .drop_duplicates("ROW")
                    .index.tolist()
                )
                panel_views = [layout.panel_params[i] for i in pidx]
                ratios["height_ratios"] = [
                    np.ptp(pv.y.range) for pv in panel_views
                ]

        if isinstance(self.space, dict):
            if len(self.space["x"]) != self.ncol:
                raise ValueError(
                    "The number of x-ratios for the facet space sizes "
                    "should match the number of columns."
                )

            if len(self.space["y"]) != self.nrow:
                raise ValueError(
                    "The number of y-ratios for the facet space sizes "
                    "should match the number of rows."
                )

            ratios["width_ratios"] = self.space.get("x")
            ratios["height_ratios"] = self.space.get("y")

        return plt.figure(), GridSpec(self.nrow, self.ncol, **ratios)

    def compute_layout(self, data: list[pd.DataFrame]) -> pd.DataFrame:
        if not self.rows and not self.cols:
            self.nrow, self.ncol = 1, 1
            return layout_null()

        base_rows = combine_vars(
            data, self.environment, self.rows, drop=self.drop
        )

        if not self.as_table:
            # Reverse the order of the rows
            base_rows = base_rows[::-1]
        base_cols = combine_vars(
            data, self.environment, self.cols, drop=self.drop
        )

        base = cross_join(base_rows, base_cols)

        if self.margins:
            base = add_margins(base, (self.rows, self.cols), self.margins)
            base = base.drop_duplicates().reset_index(drop=True)

        n = len(base)
        panel = ninteraction(base, drop=True)
        panel = pd.Categorical(panel, categories=range(1, n + 1))

        if self.rows:
            rows = ninteraction(base[self.rows], drop=True)
        else:
            rows = [1] * len(panel)

        if self.cols:
            cols = ninteraction(base[self.cols], drop=True)
        else:
            cols = [1] * len(panel)

        layout = pd.DataFrame(
            {
                "PANEL": panel,
                "ROW": rows,
                "COL": cols,
            }
        )
        layout = pd.concat([layout, base], axis=1)
        layout = layout.sort_values("PANEL")
        layout.reset_index(drop=True, inplace=True)

        # Relax constraints, if necessary
        layout["SCALE_X"] = layout["COL"] if self.free["x"] else 1
        layout["SCALE_Y"] = layout["ROW"] if self.free["y"] else 1
        layout["AXIS_X"] = layout["ROW"] == layout["ROW"].max()
        layout["AXIS_Y"] = layout["COL"] == layout["COL"].min()

        self.nrow = layout["ROW"].max()
        self.ncol = layout["COL"].max()
        return layout

    def map(self, data: pd.DataFrame, layout: pd.DataFrame) -> pd.DataFrame:
        if not len(data):
            data["PANEL"] = pd.Categorical(
                [], categories=layout["PANEL"].cat.categories, ordered=True
            )
            return data

        vars = (*self.rows, *self.cols)
        margin_vars: tuple[list[str], list[str]] = (
            list(data.columns.intersection(self.rows)),
            list(data.columns.intersection(self.cols)),
        )
        data = add_margins(data, margin_vars, self.margins)

        facet_vals = eval_facet_vars(data, vars, self.environment)
        data, facet_vals = add_missing_facets(data, layout, vars, facet_vals)

        # assign each point to a panel
        if len(facet_vals) and len(facet_vals.columns):
            keys = join_keys(facet_vals, layout, vars)
            data["PANEL"] = match(keys["x"], keys["y"], start=1)
        else:
            # Special case of no facetting
            data["PANEL"] = 1

        # matching dtype and
        # the categories(panel numbers) for the data should be in the
        # same order as the panels. i.e the panels are the reference,
        # they "know" the right order
        data["PANEL"] = pd.Categorical(
            data["PANEL"],
            categories=layout["PANEL"].cat.categories,
            ordered=True,
        )

        data.reset_index(drop=True, inplace=True)
        return data

    def make_strips(self, layout_info: layout_details, ax: Axes) -> Strips:
        lst = []
        if layout_info.is_top and self.cols:
            s = strip(self.cols, layout_info, self, ax, "top")
            lst.append(s)
        if layout_info.is_right and self.rows:
            s = strip(self.rows, layout_info, self, ax, "right")
            lst.append(s)
        return Strips(lst)


def parse_grid_rows_cols(
    rows: Optional[str | Sequence[str]] = None,
    cols: Optional[str | Sequence[str]] = None,
) -> tuple[list[str], list[str]]:
    """
    Return the rows & cols that make up the grid
    """
    if cols is None and isinstance(rows, str):  # formula
        return parse_grid_facets_old(rows)

    if cols is None:
        cols = []
    elif isinstance(cols, str):
        cols = [cols]

    if rows is None:
        rows = []
    elif isinstance(rows, str):
        rows = [rows]

    return list(rows), list(cols)


def parse_grid_facets_old(
    facets: str | tuple[str | Sequence[str], str | Sequence[str]],
) -> tuple[list[str], list[str]]:
    """
    Return two lists of facetting variables, for the rows & columns

    This parse the old & silently deprecated style.
    """
    valid_seqs = [
        "(var1,)",
        "('var1', '.')",
        "('var1', 'var2')",
        "('.', 'var1')",
        "((var1, var2), (var3, var4))",
    ]
    error_msg_s = (
        "Valid sequences for specifying 'facets' look like" f" {valid_seqs}"
    )

    valid_forms = [
        "var1",
        "var1 ~ .",
        "var1 ~ var2",
        ". ~ var1",
        "var1 + var2 ~ var3 + var4",
        ". ~ func(var1) + func(var2)",
        ". ~ func(var1+var3) + func(var2)",
    ] + valid_seqs
    error_msg_f = "Valid formula for 'facet_grid' look like" f" {valid_forms}"

    if not isinstance(facets, str):
        if len(facets) == 1:
            rows = ensure_list_spec(facets[0])
            cols = []
        elif len(facets) == 2:
            rows = ensure_list_spec(facets[0])
            cols = ensure_list_spec(facets[1])
        else:
            raise PlotnineError(error_msg_s)
        return list(rows), list(cols)

    if "~" not in facets:
        rows = ensure_list_spec(facets)
        return list(rows), []

    # Example of allowed formulae
    # "c ~ a + b'
    # '. ~ func(a) + func(b)'
    # 'func(c) ~ func(a+1) + func(b+2)'
    try:
        lhs, rhs = facets.split("~")
    except ValueError as e:
        raise PlotnineError(error_msg_f) from e
    else:
        lhs = lhs.strip()
        rhs = rhs.strip()

    rows = ensure_list_spec(lhs)
    cols = ensure_list_spec(rhs)
    return list(rows), list(cols)


def ensure_list_spec(term: Sequence[str] | str) -> Sequence[str]:
    """
    Convert a str specification to a list spec

    e.g.
    'a' -> ['a']
    'a + b' -> ['a', 'b']
    '.' -> []
    '' -> []
    """
    if isinstance(term, str):
        splitter = " + " if " + " in term else "+"
        if term in [".", ""]:
            return []
        return [var.strip() for var in term.split(splitter)]
    else:
        return term



from warnings import warn

import numpy as np

from .._utils.registry import alias
from ..doctools import document
from ..exceptions import PlotnineWarning
from .scale_continuous import scale_continuous
from .scale_discrete import scale_discrete


@document
class scale_stroke_continuous(scale_continuous):
    """
    Continuous Stroke Scale

    Parameters
    ----------
    range :
        Range ([Minimum, Maximum]) of output stroke values.
        Should be between 0 and 1.
    {superclass_parameters}
    """

    _aesthetics = ["stroke"]

    def __init__(self, range: tuple[float, float] = (1, 6), **kwargs):
        from mizani.palettes import rescale_pal

        # TODO: fix types in mizani
        self.palette = rescale_pal(range)  # pyright: ignore
        scale_continuous.__init__(self, **kwargs)


@document
class scale_stroke_ordinal(scale_discrete):
    """
    Discrete Stroke Scale

    Parameters
    ----------
    range :
        Range ([Minimum, Maximum]) of output stroke values.
        Should be between 0 and 1.
    {superclass_parameters}
    """

    _aesthetics = ["stroke"]

    def __init__(self, range: tuple[float, float] = (1, 6), **kwargs):
        def palette(value: int):
            return np.linspace(range[0], range[1], value)

        self.palette = palette
        scale_discrete.__init__(self, **kwargs)


@document
class scale_stroke_discrete(scale_stroke_ordinal):
    """
    Discrete Stroke Scale

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["stroke"]

    def __init__(self, **kwargs):
        warn(
            "Using stroke for a ordinal variable is not advised.",
            PlotnineWarning,
        )
        super().__init__(**kwargs)


@alias
class scale_stroke(scale_stroke_continuous):
    pass



import numpy as np
import pandas as pd
import pytest

from plotnine import (
    aes,
    after_stat,
    geom_col,
    geom_crossbar,
    geom_point,
    ggplot,
    scale_x_log10,
    scale_y_log10,
    stage,
    stat_bin_2d,
    stat_ecdf,
    stat_function,
)
from plotnine.mapping.aes import make_labels

data = pd.DataFrame(
    {
        "x": pd.Categorical(["b", "d", "c", "a"], ordered=True),
        "y": [1, 2, 3, 4],
    }
)


def test_reorder():
    p = (
        ggplot(data, aes("reorder(x, y)", "y", fill="reorder(x, y)"))
        + geom_col()
    )
    assert p == "reorder"


def test_reorder_index():
    # The dataframe is created with ordering according to the y
    # variable. So the x index should be ordered acc. to y too
    p = ggplot(data, aes("reorder(x, x.index)", "y")) + geom_col()
    assert p == "reorder_index"


def test_labels_series():
    p = ggplot(data, aes(x=data.x, y=data.y)) + geom_col()
    assert p.labels.x == "x"
    assert p.labels.y == "y"


def test_labels_lists():
    p = ggplot(data, aes(x=[1, 2, 3], y=[1, 2, 3])) + geom_col()
    assert p.labels.x is None
    assert p.labels.y is None


class TestTransScale:
    df1 = pd.DataFrame({"var": np.arange(1, 11) / 10})

    def test_stat_function(self):
        p = (
            ggplot(self.df1, aes(x="var"))
            + geom_point(aes(y="var"))
            + stat_function(fun=lambda x: x)
            + scale_y_log10()
        )
        assert p == "test_stat_function"

    def test_stat_ecdf(self):
        p = (
            ggplot(self.df1, aes(x="var"))
            + geom_point(aes(y="var"))
            + stat_ecdf()
            + scale_y_log10()
        )

        with pytest.warns(RuntimeWarning):
            assert p == "stat_ecdf"

    def test_stat_bin_2d(self):
        data = pd.DataFrame({"x": [1, 10, 100, 1000], "y": range(4)})
        p = ggplot(data, aes("x", "y")) + stat_bin_2d(bins=3) + scale_x_log10()
        assert p == "stat_bin_2d"

    def test_geom_crossbar(self):
        data = pd.DataFrame({"x": "x", "y": np.linspace(0.1, 1, 10)})

        p = (
            ggplot(data, aes("x", "y"))
            + geom_crossbar(
                aes(
                    ymin=after_stat("lower"),
                    y=stage(start="y", after_stat="middle"),
                    ymax=after_stat("upper"),
                ),
                stat="boxplot",
            )
            + scale_y_log10()
        )
        assert p == "geom_crossbar"


def test_make_labels():
    mapping = {"y": "y", "color": ["Treatment"]}
    labels = make_labels(mapping)
    assert labels.color == "Treatment"

    mapping = {"y": "y", "color": ["Treatment", "Control"]}
    labels = make_labels(mapping)
    assert labels.color is None



from __future__ import annotations

import typing
from contextlib import suppress
from warnings import warn

import numpy as np
import pandas as pd
from mizani.bounds import censor, expand_range_distinct, rescale, zero_range

from .._utils import match
from ..doctools import document
from ..exceptions import PlotnineError, PlotnineWarning
from ..iapi import range_view, scale_view
from ._expand import expand_range
from .range import RangeContinuous
from .scale import scale

if typing.TYPE_CHECKING:
    from typing import Any, Optional, Sequence, Type

    from mizani.transforms import trans

    from plotnine.typing import (
        CoordRange,
        FloatArrayLike,
        ScaleContinuousBreaks,
        ScaleContinuousBreaksRaw,
        ScaleContinuousLimits,
        ScaleContinuousLimitsRaw,
        ScaleLabels,
        ScaleMinorBreaksRaw,
        TFloatArrayLike,
        TupleFloat2,
        TupleFloat4,
    )


@document
class scale_continuous(scale):
    """
    Base class for all continuous scales

    Parameters
    ----------
    {superclass_parameters}
    trans : str | callable
        Name of a trans function or a trans function.
        See [](`mizani.transforms`) for possible options.
    oob : callable, default=mizani.bounds.censor
        Function to deal with out of bounds (limits)
        data points. Default is to turn them into
        `np.nan`, which then get dropped.
    minor_breaks : list | int | callable, default=None
        If a list-like, it is the minor breaks points.
        If an integer, it is the number of minor breaks between
        any set of major breaks.
        If a function, it should have the signature
        `func(limits)` and return a list-like of consisting
        of the minor break points.
        If `None`, no minor breaks are calculated.
        The default is to automatically calculate them.
    rescaler : callable, default=mizani.bounds.rescale
        Function to rescale data points so that they can
        be handled by the palette. Default is to rescale
        them onto the [0, 1] range. Scales that inherit
        from this class may have another default.

    Notes
    -----
    If using the class directly all arguments must be
    keyword arguments.
    """

    _range_class = RangeContinuous
    _limits: Optional[ScaleContinuousLimitsRaw]
    range: RangeContinuous
    rescaler = staticmethod(rescale)  # Used by diverging & n colour gradients
    oob = staticmethod(censor)  # what to do with out of bounds data points
    breaks: ScaleContinuousBreaksRaw
    minor_breaks: ScaleMinorBreaksRaw = True
    _trans: trans | str = "identity"  # transform class

    def __init__(self, **kwargs):
        # Make sure we have a transform.
        self.trans = kwargs.pop("trans", self._trans)

        with suppress(KeyError):
            self.limits = kwargs.pop("limits")

        scale.__init__(self, **kwargs)

    def _check_trans(self, t):
        """
        Check if transform t is valid

        When scales specialise on a specific transform (other than
        the identity transform), the user should know when they
        try to change the transform.

        Parameters
        ----------
        t : mizani.transforms.trans
            Transform object
        """
        orig_trans_name = self.__class__._trans
        new_trans_name = t.__class__.__name__
        if new_trans_name.endswith("_trans"):
            new_trans_name = new_trans_name[:-6]
        if orig_trans_name not in ("identity", new_trans_name):
            warn(
                "You have changed the transform of a specialised scale. "
                "The result may not be what you expect.\n"
                "Original transform: {}\n"
                "New transform: {}".format(orig_trans_name, new_trans_name),
                PlotnineWarning,
                stacklevel=2,
            )

    @property
    def trans(self) -> trans:
        return self._trans  # pyright: ignore

    @trans.setter
    def trans(self, value: trans | str | Type[trans]):
        from mizani.transforms import gettrans

        t: trans = gettrans(value)
        self._check_trans(t)
        self._trans = t

    @property
    def limits(self):
        if self.is_empty():
            return (0, 1)

        if self._limits is None:
            return self.range.range
        elif callable(self._limits):
            # Function works in the dataspace, but the limits are
            # stored in transformed space. The range of the scale is
            # in transformed space (i.e. with in the domain of the scale)
            _range = self.inverse(self.range.range)
            return self.trans.transform(self._limits(_range))
        elif (
            self._limits is not None
            and not self.range.is_empty()
            and
            # Fall back to the range if the limits
            # are not set or if any is None or NaN
            len(self._limits) == len(self.range.range)
        ):
            l1, l2 = self._limits
            r1, r2 = self.range.range
            if l1 is None:
                l1 = self.trans.transform([r1])[0]
            if l2 is None:
                l2 = self.trans.transform([r2])[0]
            return l1, l2

        return self._limits

    @limits.setter
    def limits(self, value: ScaleContinuousLimitsRaw):
        """
        Limits for the continuous scale

        Parameters
        ----------
        value : array_like | callable
            Limits in the dataspace.
        """
        # Notes
        # -----
        # The limits are given in original dataspace
        # but they are stored in transformed space since
        # all computations happen on transformed data. The
        # labeling of the plot axis and the guides are in
        # the original dataspace.
        if isinstance(value, bool) or value is None or callable(value):
            self._limits = value
            return

        a, b = value
        a = self.transform([a])[0] if a is not None else a
        b = self.transform([b])[0] if b is not None else b

        if a is not None and b is not None and a > b:
            a, b = b, a

        self._limits = a, b

    def train(self, x: FloatArrayLike):
        """
        Train continuous scale
        """
        if not len(x):
            return

        self.range.train(x)

    def transform_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transform dataframe
        """
        if len(df) == 0:
            return df

        aesthetics = set(self.aesthetics) & set(df.columns)
        for ae in aesthetics:
            with suppress(TypeError):
                df[ae] = self.transform(df[ae])

        return df

    def transform(self, x: TFloatArrayLike) -> TFloatArrayLike:
        """
        Transform array|series x
        """
        try:
            return self.trans.transform(x)
        except TypeError:
            return [self.trans.transform(val) for val in x]  # pyright: ignore

    def inverse_df(self, df):
        """
        Inverse Transform dataframe
        """
        if len(df) == 0:
            return df

        aesthetics = set(self.aesthetics) & set(df.columns)
        for ae in aesthetics:
            with suppress(TypeError):
                df[ae] = self.inverse(df[ae])

        return df

    def inverse(self, x: TFloatArrayLike) -> TFloatArrayLike:
        """
        Inverse transform array|series x
        """
        try:
            return self.trans.inverse(x)
        except TypeError:
            return [self.trans.inverse(val) for val in x]  # pyright: ignore

    @property
    def is_linear(self) -> bool:
        """
        Return True if the scale is linear

        Depends on the transformation.
        """
        return self.trans.transform_is_linear

    def dimension(self, expand=(0, 0, 0, 0), limits=None):
        """
        Get the phyical size of the scale

        Unlike limits, this always returns a numeric vector of length 2
        """
        if limits is None:
            limits = self.limits
        return expand_range_distinct(limits, expand)

    def expand_limits(
        self,
        limits: ScaleContinuousLimits,
        expand: TupleFloat2 | TupleFloat4,
        coord_limits: CoordRange | None,
        trans: trans,
    ) -> range_view:
        """
        Calculate the final range in coordinate space
        """
        # - Override None in coord_limits
        # - Expand limits in coordinate space
        # - Remove any computed infinite values &
        if coord_limits is not None:
            c0, c1 = coord_limits
            limits = (
                limits[0] if c0 is None else c0,
                limits[1] if c1 is None else c1,
            )
        return expand_range(limits, expand, trans)

    def view(
        self,
        limits: Optional[CoordRange] = None,
        range: Optional[CoordRange] = None,
    ) -> scale_view:
        """
        Information about the trained scale
        """
        if limits is None:
            limits = self.limits

        if range is None:
            range = self.dimension(limits=limits)

        breaks = self.get_bounded_breaks(range)
        labels = self.get_labels(breaks)

        ubreaks = self.get_breaks(range)
        minor_breaks = self.get_minor_breaks(ubreaks, range)

        sv = scale_view(
            scale=self,
            aesthetics=self.aesthetics,
            name=self.name,
            limits=limits,
            range=range,
            breaks=breaks,
            labels=labels,
            minor_breaks=minor_breaks,
        )
        return sv

    def default_expansion(self, mult=0.05, add=0, expand=True):
        """
        Get the default expansion for continuous scale
        """
        return super().default_expansion(mult, add, expand)

    def palette(self, value: FloatArrayLike) -> Sequence[Any]:
        """
        Aesthetic mapping function
        """
        return super().palette(value)

    def map(
        self, x: FloatArrayLike, limits: Optional[ScaleContinuousLimits] = None
    ) -> FloatArrayLike:
        if limits is None:
            limits = self.limits

        x = self.oob(self.rescaler(x, _from=limits))

        uniq = np.unique(x)
        pal = np.asarray(self.palette(uniq))
        scaled = pal[match(x, uniq)]
        if scaled.dtype.kind == "U":
            scaled = [self.na_value if x == "nan" else x for x in scaled]
        else:
            scaled[pd.isna(scaled)] = self.na_value
        return scaled

    def get_breaks(
        self, limits: Optional[ScaleContinuousLimits] = None
    ) -> ScaleContinuousBreaks:
        """
        Generate breaks for the axis or legend

        Parameters
        ----------
        limits : list_like | None
            If None the self.limits are used
            They are expected to be in transformed
            space.

        Returns
        -------
        out : array_like

        Notes
        -----
        Breaks are calculated in data space and
        returned in transformed space since all
        data is plotted in transformed space.
        """
        if limits is None:
            limits = self.limits

        # To data space
        _limits = self.inverse(limits)

        if self.is_empty() or self.breaks is False or self.breaks is None:
            breaks = []
        elif self.breaks is True:
            # TODO: Fix this type mismatch in mizani with
            # a typevar so that type-in = type-out
            _tlimits = self.trans.breaks(_limits)
            breaks: ScaleContinuousBreaks = _tlimits  # pyright: ignore
        elif zero_range(_limits):
            breaks = [_limits[0]]
        elif callable(self.breaks):
            breaks = self.breaks(_limits)
        else:
            breaks = self.breaks

        breaks = self.transform(breaks)
        return breaks

    def get_bounded_breaks(
        self, limits: Optional[ScaleContinuousLimits] = None
    ) -> ScaleContinuousBreaks:
        """
        Return Breaks that are within limits
        """
        if limits is None:
            limits = self.limits
        breaks = self.get_breaks(limits)
        strict_breaks = [b for b in breaks if limits[0] <= b <= limits[1]]
        return strict_breaks

    def get_minor_breaks(
        self,
        major: ScaleContinuousBreaks,
        limits: Optional[ScaleContinuousLimits] = None,
    ) -> ScaleContinuousBreaks:
        """
        Return minor breaks
        """
        if limits is None:
            limits = self.limits

        if self.minor_breaks is False or self.minor_breaks is None:
            minor_breaks = []
        elif self.minor_breaks is True:
            minor_breaks: ScaleContinuousBreaks = self.trans.minor_breaks(
                major, limits
            )  # pyright: ignore
        elif isinstance(self.minor_breaks, int):
            minor_breaks: ScaleContinuousBreaks = self.trans.minor_breaks(
                major,
                limits,
                self.minor_breaks,  # pyright: ignore
            )
        elif callable(self.minor_breaks):
            breaks = self.minor_breaks(self.inverse(limits))
            _major = set(major)
            minor = self.transform(breaks)
            minor_breaks = [x for x in minor if x not in _major]
        else:
            minor_breaks = self.transform(self.minor_breaks)

        return minor_breaks

    def get_labels(
        self, breaks: Optional[ScaleContinuousBreaks] = None
    ) -> ScaleLabels:
        """
        Generate labels for the axis or legend

        Parameters
        ----------
        breaks: None | array_like
            If None, use self.breaks.
        """
        if breaks is None:
            breaks = self.get_breaks()

        breaks = self.inverse(breaks)
        labels: Sequence[str]

        if self.labels is False or self.labels is None:
            labels = []
        elif self.labels is True:
            labels = self.trans.format(breaks)  # type: ignore
        elif callable(self.labels):
            labels = self.labels(breaks)
        elif isinstance(self.labels, dict):
            labels = [
                str(self.labels[b]) if b in self.labels else str(b)
                for b in breaks
            ]
        else:
            # When user sets breaks and labels of equal size,
            # but the limits exclude some of the breaks.
            # We remove the corresponding labels
            from collections.abc import Sized

            labels = self.labels
            if (
                len(labels) != len(breaks)
                and isinstance(self.breaks, Sized)
                and len(labels) == len(self.breaks)
            ):
                _wanted_breaks = set(breaks)
                labels = [
                    l
                    for l, b in zip(labels, self.breaks)
                    if b in _wanted_breaks
                ]

        if len(labels) != len(breaks):
            raise PlotnineError("Breaks and labels are different lengths")

        return labels



from __future__ import annotations

import typing
from collections.abc import Sequence
from copy import copy, deepcopy
from io import BytesIO
from itertools import chain
from pathlib import Path
from types import SimpleNamespace as NS
from typing import Any, Dict, Iterable, Optional
from warnings import warn

from ._utils import (
    from_inches,
    is_data_like,
    order_as_data_mapping,
    to_inches,
    ungroup,
)
from ._utils.context import plot_context
from ._utils.ipython import (
    get_display_function,
    get_ipython,
    is_inline_backend,
)
from .coords import coord_cartesian
from .exceptions import PlotnineError, PlotnineWarning
from .facets import facet_null
from .facets.layout import Layout
from .geoms.geom_blank import geom_blank
from .guides.guides import guides
from .iapi import mpl_save_view
from .layer import Layers
from .mapping.aes import aes, make_labels
from .options import get_option
from .scales.scales import Scales
from .themes.theme import theme, theme_get

if typing.TYPE_CHECKING:
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
    from typing_extensions import Self

    from plotnine import watermark
    from plotnine.coords.coord import coord
    from plotnine.facets.facet import facet
    from plotnine.layer import layer
    from plotnine.typing import DataLike, PlotAddable

__all__ = ("ggplot", "ggsave", "save_as_pdf_pages")


class ggplot:
    """
    Create a new ggplot object

    Parameters
    ----------
    data :
        Default data for plot. Every layer that does not
        have data of its own will use this one.
    mapping :
        Default aesthetics mapping for the plot. These will be used
        by all layers unless specifically overridden.

    Notes
    -----
    ggplot object only have partial support for pickling. The mappings used
    by pickled objects should not reference variables in the namespace.
    """

    figure: Figure
    axs: list[Axes]
    theme: theme
    facet: facet
    coordinates: coord

    def __init__(
        self,
        data: Optional[DataLike] = None,
        mapping: Optional[aes] = None,
    ):
        from .mapping._env import Environment

        # Allow some sloppiness
        data, mapping = order_as_data_mapping(data, mapping)
        self.data = data
        self.mapping = mapping if mapping is not None else aes()
        self.facet = facet_null()
        self.labels = make_labels(self.mapping)
        self.layers = Layers()
        self.guides = guides()
        self.scales = Scales()
        self.theme = theme_get()
        self.coordinates: coord = coord_cartesian()
        self.environment = Environment.capture(1)
        self.layout = Layout()
        self.watermarks: list[watermark] = []

        # build artefacts
        self._build_objs = NS()

    def __str__(self) -> str:
        """
        Print/show the plot
        """
        msg = (
            "Using print(plot) to draw and show the plot figure is "
            "deprecated and will be removed in a future version. "
            "Use plot.show()."
        )
        warn(msg, category=FutureWarning, stacklevel=2)
        self.show()
        # Return and empty string so that print(p) is as clean as possible
        return ""

    def __repr__(self) -> str:
        """
        Print/show the plot
        """
        dpi = self.theme.getp("dpi")
        width, height = self.theme.getp("figure_size")
        W, H = int(width * dpi), int(height * dpi)

        msg = (
            "Using repr(plot) to draw and show the plot figure is "
            "deprecated and will be removed in a future version. "
            "Use plot.show()."
        )
        warn(msg, category=FutureWarning, stacklevel=2)
        self.show()
        return f"<Figure Size: ({W} x {H})>"

    def _ipython_display_(self):
        """
        Display plot in the output of the cell

        This method will always be called when a ggplot object is the
        last in the cell.
        """
        self._display()

    def show(self):
        """
        Show plot using the matplotlib backend set by the user

        Users should prefer this method instead of printing or repring
        the object.
        """
        self._display() if is_inline_backend() else self.draw(show=True)

    def _display(self):
        """
        Display plot in the cells output

        This function is called for its side-effects.

        It plots the plot to an io buffer, then uses ipython display
        methods to show the result
        """
        ip = get_ipython()
        format = get_option("figure_format") or ip.config.InlineBackend.get(
            "figure_format", "retina"
        )
        save_format = format

        # While jpegs can be displayed as retina, we restrict the output
        # of "retina" to png
        if format == "retina":
            self = copy(self)
            self.theme = self.theme.to_retina()
            save_format = "png"

        buf = BytesIO()
        self.save(buf, format=save_format, verbose=False)
        display_func = get_display_function(format)
        display_func(buf.getvalue())

    def __deepcopy__(self, memo: dict[Any, Any]) -> ggplot:
        """
        Deep copy without copying the dataframe and environment
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        # don't make a deepcopy of data
        shallow = {"data", "figure", "_build_objs"}
        for key, item in old.items():
            if key in shallow:
                new[key] = item
                memo[id(new[key])] = new[key]
            else:
                new[key] = deepcopy(item, memo)

        return result

    def __iadd__(self, other: PlotAddable | list[PlotAddable] | None) -> Self:
        """
        Add other to ggplot object

        Parameters
        ----------
        other :
            Either an object that knows how to "radd"
            itself to a ggplot, or a list of such objects.
        """
        if isinstance(other, Sequence):
            for item in other:
                item.__radd__(self)
        elif other is not None:
            other.__radd__(self)
        return self

    def __add__(self, other: PlotAddable | list[PlotAddable] | None) -> ggplot:
        """
        Add to ggplot from a list

        Parameters
        ----------
        other :
            Either an object that knows how to "radd"
            itself to a ggplot, or a list of such objects.
        """
        self = deepcopy(self)
        return self.__iadd__(other)

    def __rrshift__(self, other: DataLike) -> ggplot:
        """
        Overload the >> operator to receive a dataframe
        """
        other = ungroup(other)
        if is_data_like(other):
            if self.data is None:
                self.data = other
            else:
                raise PlotnineError("`>>` failed, ggplot object has data.")
        else:
            msg = "Unknown type of data -- {!r}"
            raise TypeError(msg.format(type(other)))
        return self

    def draw(self, show: bool = False) -> Figure:
        """
        Render the complete plot

        Parameters
        ----------
        show :
            Whether to show the plot.

        Returns
        -------
        :
            Matplotlib figure
        """
        from ._mpl.layout_engine import PlotnineLayoutEngine

        # Do not draw if drawn already.
        # This prevents a needless error when reusing
        # figure & axes in the jupyter notebook.
        if hasattr(self, "figure"):
            return self.figure

        # Prevent against any modifications to the users
        # ggplot object. Do the copy here as we may/may not
        # assign a default theme
        self = deepcopy(self)
        with plot_context(self, show=show):
            self._build()

            # setup
            self.figure, self.axs = self.facet.setup(self)
            self.guides._setup(self)
            self.theme.setup(self)

            # Drawing
            self._draw_layers()
            self._draw_panel_borders()
            self._draw_breaks_and_labels()
            self.guides.draw()
            self._draw_figure_texts()
            self._draw_watermarks()

            # Artist object theming
            self.theme.apply()
            self.figure.set_layout_engine(PlotnineLayoutEngine(self))

        return self.figure

    def _draw_using_figure(self, figure: Figure, axs: list[Axes]) -> ggplot:
        """
        Draw onto already created figure and axes

        This is can be used to draw animation frames,
        or inset plots. It is intended to be used
        after the key plot has been drawn.

        Parameters
        ----------
        figure :
            Matplotlib figure
        axs :
            Array of Axes onto which to draw the plots
        """
        from ._mpl.layout_engine import PlotnineLayoutEngine

        self = deepcopy(self)
        self.figure = figure
        self.axs = axs
        with plot_context(self):
            self._build()

            # setup
            self.figure, self.axs = self.facet.setup(self)
            self.guides._setup(self)
            self.theme.setup(self)

            # drawing
            self._draw_layers()
            self._draw_breaks_and_labels()
            self.guides.draw()

            # artist theming
            self.theme.apply()
            self.figure.set_layout_engine(PlotnineLayoutEngine(self))

        return self

    def _build(self):
        """
        Build ggplot for rendering.

        Notes
        -----
        This method modifies the ggplot object. The caller is
        responsible for making a copy and using that to make
        the method call.
        """
        if not self.layers:
            self += geom_blank()

        layers = self._build_objs.layers = self.layers
        scales = self._build_objs.scales = self.scales
        layout = self._build_objs.layout = self.layout

        # Update the label information for the plot
        layers.update_labels(self)

        # Give each layer a copy of the data, the mappings and
        # the execution environment
        layers.setup(self)

        # Initialise panels, add extra data for margins & missing
        # facetting variables, and add on a PANEL variable to data
        layout.setup(layers, self)

        # Compute aesthetics to produce data with generalised
        # variable names
        layers.compute_aesthetics(self)

        # Transform data using all scales
        layers.transform(scales)

        # Make sure missing (but required) aesthetics are added
        scales.add_missing(("x", "y"))

        # Map and train positions so that statistics have access
        # to ranges and all positions are numeric
        layout.train_position(layers, scales)
        layout.map_position(layers)

        # Apply and map statistics
        layers.compute_statistic(layout)
        layers.map_statistic(self)

        # Prepare data in geoms
        # e.g. from y and width to ymin and ymax
        layers.setup_data()

        # Apply position adjustments
        layers.compute_position(layout)

        # Reset position scales, then re-train and map.  This
        # ensures that facets have control over the range of
        # a plot.
        layout.reset_position_scales()
        layout.train_position(layers, scales)
        layout.map_position(layers)

        # Train and map non-position scales
        npscales = scales.non_position_scales()
        if len(npscales):
            layers.train(npscales)
            layers.map(npscales)

        # Train coordinate system
        layout.setup_panel_params(self.coordinates)

        # fill in the defaults
        layers.use_defaults_after_scale()

        # Allow stats to modify the layer data
        layers.finish_statistics()

        # Allow layout to modify data before rendering
        layout.finish_data(layers)

    def _draw_panel_borders(self):
        """
        Draw Panel boders
        """
        # We add a patch rather than use ax.patch because want the
        # grid lines below the borders. We leave ax.patch for the
        # background only.
        if self.theme.T.is_blank("panel_border"):
            return

        from matplotlib.patches import Rectangle

        for ax in self.axs:
            rect = Rectangle(
                (0, 0),
                1,
                1,
                facecolor="none",
                transform=ax.transAxes,
                # Adding a clip path but defaulting to no clipping
                # gives a fullwidth border that can perfectly overlap
                # will with legend borders.
                clip_path=ax.patch,
                clip_on=False,
            )
            self.figure.add_artist(rect)
            self.theme.targets.panel_border.append(rect)

    def _draw_layers(self):
        """
        Draw the main plot(s) onto the axes.
        """
        # Draw the geoms
        self.layers.draw(self.layout, self.coordinates)

    def _draw_breaks_and_labels(self):
        """
        Draw breaks and labels
        """
        # 1. Draw facet labels a.k.a strip text
        # 2. Decorate the axes
        #      - xaxis & yaxis breaks, labels, limits, ...
        #
        # pidx is the panel index (location left to right, top to bottom)
        self.facet.strips.draw()
        for layout_info in self.layout.get_details():
            pidx = layout_info.panel_index
            ax = self.axs[pidx]
            panel_params = self.layout.panel_params[pidx]
            self.facet.set_limits_breaks_and_labels(panel_params, ax)

            # Remove unnecessary ticks and labels
            if not layout_info.axis_x:
                ax.xaxis.set_tick_params(
                    which="both", bottom=False, labelbottom=False
                )
            if not layout_info.axis_y:
                ax.yaxis.set_tick_params(
                    which="both", left=False, labelleft=False
                )

            if layout_info.axis_x:
                ax.xaxis.set_tick_params(which="both", bottom=True)
            if layout_info.axis_y:
                ax.yaxis.set_tick_params(which="both", left=True)

    def _draw_figure_texts(self):
        """
        Draw title, x label, y label and caption onto the figure
        """
        figure = self.figure
        theme = self.theme
        targets = theme.targets

        title = self.labels.get("title", "")
        subtitle = self.labels.get("subtitle", "")
        caption = self.labels.get("caption", "")

        # Get the axis labels (default or specified by user)
        # and let the coordinate modify them e.g. flip
        labels = self.coordinates.labels(
            self.layout.set_xy_labels(self.labels)
        )

        # The locations are handled by the layout manager
        if title:
            targets.plot_title = figure.text(0, 0, title)

        if subtitle:
            targets.plot_subtitle = figure.text(0, 0, subtitle)

        if caption:
            targets.plot_caption = figure.text(0, 0, caption)

        if labels.x:
            targets.axis_title_x = figure.text(0, 0, labels.x)

        if labels.y:
            targets.axis_title_y = figure.text(0, 0, labels.y)

    def _draw_watermarks(self):
        """
        Draw watermark onto figure
        """
        for wm in self.watermarks:
            wm.draw(self.figure)

    def _save_filename(self, ext: str) -> Path:
        """
        Make a filename for use by the save method

        Parameters
        ----------
        ext : str
            Extension e.g. png, pdf, ...
        """
        hash_token = abs(self.__hash__())
        return Path(f"plotnine-save-{hash_token}.{ext}")

    def _update_labels(self, layer: layer):
        """
        Update label data for the ggplot

        Parameters
        ----------
        layer : layer
            New layer that has just been added to the ggplot
            object.
        """
        mapping = make_labels(layer.mapping)
        default = make_labels(layer.stat.DEFAULT_AES)
        mapping.add_defaults(default)
        self.labels.add_defaults(mapping)

    def save_helper(
        self: ggplot,
        filename: Optional[str | Path | BytesIO] = None,
        format: Optional[str] = None,
        path: Optional[str] = None,
        width: Optional[float] = None,
        height: Optional[float] = None,
        units: str = "in",
        dpi: Optional[float] = None,
        limitsize: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> mpl_save_view:
        """
        Create MPL figure that will be saved

        Notes
        -----
        This method has the same arguments as [](`~plotnine.ggplot.save`).
        Use it to get access to the figure that will be saved.
        """
        fig_kwargs: Dict[str, Any] = {"format": format, **kwargs}

        # filename, depends on the object
        if filename is None:
            ext = format if format else "pdf"
            filename = self._save_filename(ext)

        if path and isinstance(filename, (Path, str)):
            filename = Path(path) / filename

        fig_kwargs["fname"] = filename

        # Preserve the users object
        self = deepcopy(self)

        # The figure size should be known by the theme
        if width is not None and height is not None:
            width = to_inches(width, units)
            height = to_inches(height, units)
            self += theme(figure_size=(width, height))
        elif (width is None and height is not None) or (
            width is not None and height is None
        ):
            raise PlotnineError("You must specify both width and height")

        width, height = self.theme.getp("figure_size")
        assert width is not None
        assert height is not None

        if limitsize and (width > 25 or height > 25):
            raise PlotnineError(
                f"Dimensions ({width=}, {height=}) exceed 25 inches "
                "(height and width are specified in inches/cm/mm, "
                "not pixels). If you are sure you want these "
                "dimensions, use 'limitsize=False'."
            )

        if verbose:
            _w = from_inches(width, units)
            _h = from_inches(height, units)
            warn(f"Saving {_w} x {_h} {units} image.", PlotnineWarning)
            warn(f"Filename: {filename}", PlotnineWarning)

        if dpi is not None:
            self.theme = self.theme + theme(dpi=dpi)

        figure = self.draw(show=False)
        return mpl_save_view(figure, fig_kwargs)

    def save(
        self,
        filename: Optional[str | Path | BytesIO] = None,
        format: Optional[str] = None,
        path: str = "",
        width: Optional[float] = None,
        height: Optional[float] = None,
        units: str = "in",
        dpi: Optional[int] = None,
        limitsize: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ):
        """
        Save a ggplot object as an image file

        Parameters
        ----------
        filename :
            File name to write the plot to. If not specified, a name
            like “plotnine-save-<hash>.<format>” is used.
        format :
            Image format to use, automatically extract from
            file name extension.
        path :
            Path to save plot to (if you just want to set path and
            not filename).
        width :
            Width (defaults to value set by the theme). If specified
            the `height` must also be given.
        height :
            Height (defaults to value set by the theme). If specified
            the `width` must also be given.
        units :
            Units for width and height when either one is explicitly
            specified (in, cm, or mm).
        dpi :
            DPI to use for raster graphics. If None, defaults to using
            the `dpi` of theme, if none is set then a `dpi` of 100.
        limitsize :
            If `True` (the default), ggsave will not save images
            larger than 50x50 inches, to prevent the common error
            of specifying dimensions in pixels.
        verbose :
            If `True`, print the saving information.
        kwargs :
            Additional arguments to pass to matplotlib `savefig()`.
        """
        sv = self.save_helper(
            filename=filename,
            format=format,
            path=path,
            width=width,
            height=height,
            units=units,
            dpi=dpi,
            limitsize=limitsize,
            verbose=verbose,
            **kwargs,
        )

        with plot_context(self).rc_context:
            sv.figure.savefig(**sv.kwargs)


ggsave = ggplot.save


def save_as_pdf_pages(
    plots: Iterable[ggplot],
    filename: Optional[str | Path] = None,
    path: str | None = None,
    verbose: bool = True,
    **kwargs: Any,
):
    """
    Save multiple [](`~plotnine.ggplot`) objects to a PDF file, one per page.

    Parameters
    ----------
    plots :
        Plot objects to write to file. `plots` may be either a
        collection such as a [](:class:`list`) or [](:class:`set`)

        ```python
        base_plot = ggplot(…)
        plots = [base_plot + ggtitle('%d of 3' % i) for i in range(1, 3)]
        save_as_pdf_pages(plots)
        ```

        or, a generator that yields [](`~plotnine.ggplot`) objects:

        ```python
        def myplots():
            for i in range(1, 3):
                yield ggplot(…) + ggtitle('%d of 3' % i)
        save_as_pdf_pages(myplots())
        ```
    filename :
        File name to write the plot to. If not specified, a name
        like “plotnine-save-<hash>.pdf” is used.
    path :
        Path to save plot to (if you just want to set path and
        not filename).
    verbose :
        If `True`, print the saving information.
    kwargs :
        Additional arguments to pass to
        [](:meth:`~matplotlib.figure.Figure.savefig`).

    Notes
    -----
    Using pandas [](:meth:`~pandas.DataFrame.groupby`) methods, tidy data
    can be "faceted" across pages:

    ```python
    from plotnine.data import mtcars

    def facet_pages(column)
        base_plot = [
            aes(x="wt", y="mpg", label="name"),
            geom_text(),
        ]
        for label, group_data in mtcars.groupby(column):
            yield ggplot(group_data) + base_plot + ggtitle(label)

    save_as_pdf_pages(facet_pages('cyl'))
    ```

    Unlike [](:meth:`~plotnine.ggplot.save`),
    [](:meth:`~plotnine.save_as_pdf_pages`)
    does not process arguments for `height` or `width`. To set the figure
    size, add [](`~plotnine.themes.themeable.figure_size`) to the theme
    for some or all of the objects in `plots`:

    ```python
    plot = ggplot(…)
    # The following are equivalent
    plot.save('filename.pdf', height=6, width=8)
    save_as_pdf_pages([plot + theme(figure_size=(8, 6))])
    ```
    """
    from matplotlib.backends.backend_pdf import PdfPages

    # as in ggplot.save()
    fig_kwargs = {"bbox_inches": "tight"}
    fig_kwargs.update(kwargs)

    # If plots is already an iterator, this is a no-op; otherwise
    # convert a list, etc. to an iterator
    plots = iter(plots)

    # filename, depends on the object
    if filename is None:
        # Take the first element from the iterator, store it, and
        # use it to generate a file name
        peek = [next(plots)]
        plots = chain(peek, plots)
        filename = peek[0]._save_filename("pdf")

    if path:
        filename = Path(path) / filename

    if verbose:
        warn(f"Filename: {filename}", PlotnineWarning)

    with PdfPages(filename, keep_empty=False) as pdf:
        # Re-add the first element to the iterator, if it was removed
        for plot in plots:
            fig = plot.draw()
            with plot_context(plot).rc_context:
                # Save as a page in the PDF file
                pdf.savefig(fig, **fig_kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

from .element_base import element_base

if TYPE_CHECKING:
    from typing import Any, Literal, Optional, Sequence

    from plotnine.typing import TupleFloat3, TupleFloat4


class element_line(element_base):
    """
    theme element: line

    used for backgrounds and borders

    parameters
    ----------
    color : str | tuple
        line color
    colour : str | tuple
        alias of color
    linetype : str | tuple
        line style. if a string, it should be one of *solid*, *dashed*,
        *dashdot* or *dotted*. you can create interesting dashed patterns
        using tuples, see [](`~matplotlib.lines.line2D.set_linestyle`).
    size : float
        line thickness
    kwargs : dict
        Parameters recognised by [](`~matplotlib.lines.line2d`).
    """

    def __init__(
        self,
        *,
        color: Optional[str | TupleFloat3 | TupleFloat4] = None,
        size: Optional[float] = None,
        linetype: Optional[str | Sequence[int]] = None,
        lineend: Optional[Literal["butt", "projecting", "round"]] = None,
        colour: Optional[str | TupleFloat3 | TupleFloat4] = None,
        **kwargs: Any,
    ):
        super().__init__()
        self.properties.update(**kwargs)

        color = color if color else colour
        if color:
            self.properties["color"] = color
        if size:
            self.properties["linewidth"] = size
        if linetype:
            self.properties["linestyle"] = linetype

        if linetype in ("solid", "-") and lineend:
            self.properties["solid_capstyle"] = lineend
        elif linetype and lineend:
            self.properties["dash_capstyle"] = lineend



from __future__ import annotations

import typing
from dataclasses import asdict, dataclass

from matplotlib.layout_engine import LayoutEngine
from matplotlib.text import Text

if typing.TYPE_CHECKING:
    from typing import Any, Optional

    from matplotlib.axes import Axes
    from matplotlib.backend_bases import RendererBase
    from matplotlib.figure import Figure

    from plotnine import ggplot, theme
    from plotnine._mpl.text import StripText
    from plotnine.facets.facet import facet
    from plotnine.iapi import legend_artists


@dataclass
class LayoutPack:
    """
    Objects required to compute the layout
    """

    axs: list[Axes]
    figure: Figure
    renderer: RendererBase
    theme: theme
    facet: facet
    axis_title_x: Optional[Text] = None
    axis_title_y: Optional[Text] = None
    # The legends references the structure that contains the
    # AnchoredOffsetboxes (groups of legends)
    legends: Optional[legend_artists] = None
    plot_caption: Optional[Text] = None
    plot_subtitle: Optional[Text] = None
    plot_title: Optional[Text] = None
    strip_text_x: Optional[list[StripText]] = None
    strip_text_y: Optional[list[StripText]] = None


class PlotnineLayoutEngine(LayoutEngine):
    """
    Implement geometry management for plotnine plots

    This layout manager automatically adjusts the location of
    objects placed around the plot panels and the subplot
    spacing parameters so that the plot fits cleanly within
    the figure area.
    """

    _adjust_compatible = True
    _colorbar_gridspec = False

    def __init__(self, plot: ggplot):
        self.plot = plot
        self.theme = plot.theme

    def execute(self, fig: Figure):
        from contextlib import nullcontext

        from ._plotnine_tight_layout import (
            get_plotnine_tight_layout,
            set_figure_artist_positions,
        )

        pack = self.setup()

        with getattr(pack.renderer, "_draw_disabled", nullcontext)():
            tparams = get_plotnine_tight_layout(pack)

        fig.subplots_adjust(**asdict(tparams.params))
        set_figure_artist_positions(pack, tparams)

    def setup(self) -> LayoutPack:
        """
        Put together objects required to do the layout
        """
        targets = self.theme.targets

        def get_target(name: str) -> Any:
            """
            Return themeable target or None
            """
            if self.theme.T.is_blank(name):
                return None
            else:
                t = getattr(targets, name)
                if isinstance(t, Text) and t.get_text() == "":
                    return None
                return t

        return LayoutPack(
            axs=self.plot.axs,
            figure=self.plot.figure,
            renderer=self.plot.figure._get_renderer(),  # pyright: ignore
            theme=self.theme,
            facet=self.plot.facet,
            axis_title_x=get_target("axis_title_x"),
            axis_title_y=get_target("axis_title_y"),
            legends=get_target("legends"),
            plot_caption=get_target("plot_caption"),
            plot_subtitle=get_target("plot_subtitle"),
            plot_title=get_target("plot_title"),
            strip_text_x=get_target("strip_text_x"),
            strip_text_y=get_target("strip_text_y"),
        )



"""
Theme elements used to decorate the graph.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .element_base import element_base

if TYPE_CHECKING:
    from typing import Any, Optional, Sequence

    from plotnine.typing import TupleFloat3, TupleFloat4


class element_rect(element_base):
    """
    Theme element: Rectangle

    Used for backgrounds and borders

    Parameters
    ----------
    fill : str | tuple
        Rectangle background color
    color : str | tuple
        Line color
    colour : str | tuple
        Alias of color
    size : float
        Line thickness
    kwargs : dict
        Parameters recognised by
        [](`~matplotlib.patches.Rectangle`).
        In some cases you can use the fancy parameters from
        [](`~matplotlib.patches.FancyBboxPatch`).
    """

    def __init__(
        self,
        fill: Optional[str | TupleFloat3 | TupleFloat4] = None,
        color: Optional[str | TupleFloat3 | TupleFloat4] = None,
        size: Optional[float] = None,
        linetype: Optional[str | Sequence[int]] = None,
        colour: Optional[str | TupleFloat3 | TupleFloat4] = None,
        **kwargs: Any,
    ):
        super().__init__()
        self.properties.update(**kwargs)

        color = color if color else colour
        if fill:
            self.properties["facecolor"] = fill
        if color:
            self.properties["edgecolor"] = color
        if size is not None:
            self.properties["linewidth"] = size
        if linetype:
            self.properties["linestyle"] = linetype



from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from typing_extensions import Self

    from plotnine import ggplot


class plot_context:
    """
    Context to setup the environment within with the plot is built

    Parameters
    ----------
    plot :
        ggplot object to be built within the context.
        exits.
    show :
        Whether to show the plot.
    """

    def __init__(self, plot: ggplot, show: bool = False):
        import matplotlib as mpl

        self.plot = plot
        self.show = show

        # Contexts
        self.rc_context = mpl.rc_context(plot.theme.rcParams)
        # Pandas deprecated is_copy, and when we create new dataframes
        # from slices we do not want complaints. We always uses the
        # new frames knowing that they are separate from the original.
        self.pd_option_context = pd.option_context(
            "mode.chained_assignment",
            None,
            "mode.copy_on_write",
            False,
        )

    def __enter__(self) -> Self:
        """
        Enclose in matplolib & pandas environments
        """
        self.rc_context.__enter__()
        self.pd_option_context.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Exit matplotlib & pandas environments
        """
        import matplotlib.pyplot as plt

        if exc_type is None:
            if self.show:
                plt.show()
            else:
                plt.close(self.plot.figure)
        else:
            # There is an exception, close any figure
            if hasattr(self.plot, "figure"):
                plt.close(self.plot.figure)

        self.rc_context.__exit__(exc_type, exc_value, exc_traceback)
        self.pd_option_context.__exit__(exc_type, exc_value, exc_traceback)



from __future__ import annotations

from collections.abc import MutableMapping
from contextlib import suppress
from dataclasses import dataclass
from functools import lru_cache
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any, Hashable, Protocol, Self

    class SupportsGetItem(Protocol):
        """
        Supports __getitem__
        """

        def __getitem__(self, key: str, /) -> Any: ...

        def __iter__(self) -> Iterator[Hashable]: ...


__all__ = ("Environment",)


@dataclass
class Environment:
    """
    A Python execution environment.

    Encapsulates a namespace for variable lookup
    """

    namespaces: list[SupportsGetItem]

    def __post_init__(self):
        self.namespace = StackedLookup(self.namespaces)
        """Where to look up variables from the encapsulated environment"""

    def with_outer_namespace(self, outer_namespace):
        """
        Return a new Environment with an extra namespace added.

        This namespace will be used only for variables that are not found
        in any existing namespace, i.e., it is "outside" them all.
        """
        return Environment(self.namespaces + [outer_namespace])

    def eval(self, expr: str, inner_namespace: SupportsGetItem = {}):
        """
        Evaluate some Python code in the encapsulated environment.

        Parameters
        ----------

        expr :
            A string containing a Python expression.

        inner_namespace :
            A dict-like object that will be checked first when `expr` attempts
            to access any variables.

        Returns
        -------
        :
            The value of `expr`.
        """
        code = _compile_eval(expr)
        return eval(
            code, {}, StackedLookup([inner_namespace] + self.namespaces)
        )

    @classmethod
    def capture(cls, eval_env: int | Self = 0):
        """
        Capture an execution environment from the stack.

        Parameters
        ----------
        eval_env :
            If `eval_env` is already an `Environment`, it is
            returned unchanged.

            Otherwise, we walk up the stack by `eval_env` steps and capture
            that function's evaluation environment.

        """
        import inspect

        if isinstance(eval_env, Environment):
            return eval_env

        frame = inspect.currentframe()
        frame_msg = "call-stack is not that deep!"
        if frame is None:
            raise ValueError(frame_msg)

        try:
            for i in range(eval_env + 1):
                frame = frame.f_back
                if frame is None:
                    raise ValueError(frame_msg)
            return cls([frame.f_locals, frame.f_globals])

        # The try/finally is important to avoid a potential reference cycle --
        # any exception traceback will carry a reference to *our* frame, which
        # contains a reference to our local variables, which would otherwise
        # carry a reference to some parent frame, where the exception was
        # caught...:
        finally:
            del frame

    def _namespace_ids(self):
        return [id(n) for n in self.namespaces]

    def __eq__(self, other):
        return (
            isinstance(other, Environment)
            and self._namespace_ids() == other._namespace_ids()
        )

    def __hash__(self):
        return hash((Environment, tuple(self._namespace_ids())))

    def __getstate__(*args, **kwargs):
        """
        Return state with no namespaces
        """
        return {"namespaces": [], "namespace": StackedLookup([])}

    def __deepcopy__(self, memo: dict[Any, Any]) -> Environment:
        """
        Shallow copy
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        for key, item in old.items():
            new[key] = item
            memo[id(new[key])] = new[key]

        return result


@dataclass
class StackedLookup(MutableMapping):
    """
    Iterative lookup in a stack of dicts

    Assignments go into an internal dict that is also the first place
    where a lookup is done.
    """

    stack: list[SupportsGetItem]

    def __post_init__(self):
        self._dict = {}
        self.stack = [self._dict] + list(self.stack)

    def __getitem__(self, key):
        for d in self.stack:
            with suppress(KeyError):
                return d[key]
        raise KeyError(key)

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __delitem__(self, key):
        del self._dict[key]

    def __contains__(self, key):
        with suppress(KeyError):
            self[key]
            return True
        return False

    def __iter__(self):
        """
        Unique keys in stacking order
        """
        return iter({key: None for d in self.stack for key in d})

    def __len__(self):
        """
        Number of unique keys
        """
        return len({key for d in self.stack for key in d})

    def get(self, key: str, default: Any = None) -> Any:
        with suppress(KeyError):
            return self[key]
        return default

    def __repr__(self):
        return f"{self.__class__.__name__}({self.stack})"

    def __getstate__(*args, **kwargs):
        """
        Return state with no namespace
        """
        d = {}
        return {"stack": [d], "_dict": d}

    def copy(self):
        return self

    def __deepcopy__(self, memo: dict[Any, Any]) -> StackedLookup:
        """
        Shallow copy
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        for key, item in old.items():
            new[key] = item
            memo[id(new[key])] = new[key]

        return result


@lru_cache(maxsize=256)
def _compile_eval(source):
    """
    Cached compile in eval mode
    """
    return compile(source, "<string-expression>", "eval")



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable, Literal, TypeAlias

    from IPython.core.interactiveshell import InteractiveShell

    FigureFormat: TypeAlias = Literal[
        "png", "retina", "jpeg", "jpg", "svg", "pdf"
    ]


def get_ipython() -> "InteractiveShell":
    """
    Return running IPython instance or None
    """
    try:
        from IPython.core.getipython import get_ipython as _get_ipython
    except ImportError as err:
        raise type(err)("IPython is has not been installed.") from err

    ip = _get_ipython()
    if ip is None:
        raise RuntimeError("Not running in a juptyer session.")

    return ip


def is_inline_backend():
    """
    Return True if the inline_backend is on

    This can only be True if also running in an jupyter/ipython session.
    """
    import matplotlib as mpl

    return "matplotlib_inline.backend_inline" in mpl.get_backend()


def get_display_function(format: FigureFormat) -> Callable[[bytes], None]:
    """
    Return a function that will display the plot image
    """
    from IPython.display import (
        SVG,
        Image,
        display_jpeg,
        display_pdf,
        display_png,
        display_svg,
    )

    def png(b: bytes):
        display_png(Image(b, format="png"))

    def retina(b: bytes):
        display_png(Image(b, format="png", retina=True))

    def jpeg(b: bytes):
        display_jpeg(Image(b, format="jpeg"))

    def svg(b: bytes):
        display_svg(SVG(b))

    def pdf(b: bytes):
        display_pdf(b, raw=True)

    lookup = {
        "png": png,
        "retina": retina,
        "jpeg": jpeg,
        "jpg": jpeg,
        "svg": svg,
        "pdf": pdf,
    }
    return lookup[format]



import numpy as np
import pandas as pd

from plotnine import (
    aes,
    element_blank,
    element_line,
    element_rect,
    element_text,
    facet_grid,
    facet_wrap,
    geom_blank,
    geom_point,
    ggplot,
    guide_colorbar,
    guide_legend,
    guides,
    labs,
    scale_y_continuous,
    theme,
)
from plotnine.data import mtcars

black_frame = theme(
    legend_frame=element_rect(color="black"),
    legend_ticks=element_line(color="black"),
)

red_frame = theme(
    legend_frame=element_rect(color="red"),
    legend_ticks=element_line(color="red"),
)


class TestLayout:
    g = (
        ggplot(mtcars, aes(x="wt", y="mpg", color="factor(gear)"))
        + geom_point()
        + labs(  # New
            x="Weight",
            y="Miles Per Gallon",
            title="Relationship between Weight and Fuel Efficiency in Cars",
            subtitle="Should we be driving lighter cars?",
            caption=(
                "The plot shows a negative correlation between car weight\n"
                "and fuel efficiency, with lighter cars generally achieving\n"
                "higher miles per gallon"
            ),
        )
    )

    def test_default(self):
        assert self.g == "default"

    def test_axis_title_x_justification(self):
        p = self.g + theme(axis_title_x=element_text(ha=0.2))
        assert p == "axis_title_x_justification"

    def test_axis_title_y_justification(self):
        p = self.g + theme(axis_title_y=element_text(va=0.8))
        assert p == "axis_title_y_justification"

    def test_plot_title_justification(self):
        p = self.g + theme(plot_title=element_text(ha=1))
        assert p == "plot_title_justification"

    def test_legend_on_top(self):
        p = self.g + theme(legend_position="top")
        assert p == "legend_at_top"

    def test_legend_on_the_left(self):
        p = self.g + theme(legend_position="left")
        assert p == "legend_on_the_left"

    def test_legend_at_the_bottom(self):
        p = self.g + theme(legend_position="bottom")
        assert p == "legend_at_the_bottom"

    def test_turn_off_guide(self):
        p1 = self.g + theme(legend_position="none")
        p2 = self.g + guides(color="none")
        p3 = self.g + guides(color=False)

        assert p1 == "turn_off_guide"
        assert p2 == "turn_off_guide"
        assert p3 == "turn_off_guide"

    def test_legends_in_different_positions(self):
        p = (
            self.g
            + aes(color="gear", fill="am", shape="factor(cyl)", alpha="vs")
            + guides(
                shape=guide_legend(position="bottom"),
                color=guide_legend(position="left"),
                alpha=guide_legend(position="left"),
            )
        )

        assert p == "legends_in_different_positions"

    def test_facet_grid(self):
        p = self.g + facet_grid("am", "gear")
        assert p == "facet_grid"

    def test_facet_wrap(self):
        p = self.g + facet_wrap("carb", nrow=2)
        assert p == "facet_wrap"

    def test_facet_wrap_scales_free(self):
        p = self.g + facet_wrap("carb", scales="free")
        assert p == "facet_wrap_scales_free"

    def test_plot_margin_aspect_ratio(self):
        # The margin should be exact in both directions even if
        # the figure has an aspect ratio != 1.
        p = (
            ggplot()
            + geom_blank()
            + theme(plot_margin=0.025, figure_size=(4, 3))
        )
        assert p == "plot_margin_aspect_ratio"

    def test_plot_margin_protruding_axis_text(self):
        data = pd.DataFrame({"x": np.arange(5), "y": np.arange(5) - 0.2})

        p = (
            ggplot(data, aes("x", "y"))
            + geom_point()
            + scale_y_continuous(labels="0 1 2 3 four-four-four-four".split())
            + labs(title="Protruding Axis Text")
            + theme(
                axis_text_y=element_text(
                    rotation=(0, 0, 0, 0, 90),
                    color=("black", "black", "black", "black", "red"),
                    va="center",
                ),
            )
        )
        assert p == "plot_margin_protruding_axis_text"

    def test_colorbar_frame(self):
        p = self.g + aes(color="gear") + black_frame
        assert p == "colorbar_frame"

    def test_different_colorbar_themes(self):
        p = (
            self.g
            + aes(color="gear", fill="am")
            + guides(
                color=guide_colorbar(theme=black_frame),
                fill=guide_colorbar(theme=red_frame),
            )
        )

        assert p == "different_colorbar_themes"


class TestLegendPositioning:
    g = ggplot(mtcars, aes(x="wt", y="mpg", color="gear")) + geom_point()

    def test_outside_legend_right_bottom(self):
        p = self.g + theme(legend_justification=0)
        assert p == "outside_legend_right_bottom"

    def test_outside_legend_left_top(self):
        p = self.g + theme(
            legend_position="left",
            legend_justification=1,
        )
        assert p == "outside_legend_left_top"

    def test_outside_legend_bottom_left(self):
        p = self.g + theme(
            legend_position="bottom",
            legend_justification=0,
        )
        assert p == "outside_legend_bottom_left"

    def test_outside_legend_top_right(self):
        p = self.g + theme(
            legend_position="top",
            legend_justification=1,
        )
        assert p == "outside_legend_top_right"

    def test_inside_legend_left(self):
        p = self.g + theme(
            legend_position="inside",
            legend_justification="left",
        )
        assert p == "inside_legend_left"

    def test_inside_legend_left2(self):
        p = self.g + theme(
            legend_position="inside",
            legend_justification=(0, 0.5),
        )
        assert p == "inside_legend_left"

    def test_inside_legend_top(self):
        p = self.g + theme(
            legend_position="inside",
            legend_justification="top",
        )
        assert p == "inside_legend_top"

    def test_inside_legend_top2(self):
        p = self.g + theme(legend_position=(0.5, 1))
        assert p == "inside_legend_top"

    def test_inside_legend_top_right(self):
        p = self.g + theme(
            legend_position="inside",
            legend_justification=(1, 1),
        )
        assert p == "inside_legend_top_right"

    def test_inside_legend_top_right2(self):
        p = self.g + theme(
            legend_position="inside",
            legend_position_inside=(1, 1),
        )
        assert p == "inside_legend_top_right"

    def test_inside_legend_top_right3(self):
        p = (
            self.g
            + guides(color=guide_colorbar(position="inside"))
            + theme(legend_position_inside=(1, 1))
        )
        assert p == "inside_legend_top_right"

    def test_inside_legend_top_right4(self):
        p = self.g + guides(color=guide_colorbar(position=(1, 1)))
        assert p == "inside_legend_top_right"

    def test_inside_legend_90pct_top_right(self):
        # Use an equal aspect ratio to easily check that the
        # top-right tip of the legend is at 90% along both dimensions
        p = self.g + theme(
            aspect_ratio=1,
            legend_position=(0.9, 0.9),
            legend_justification=(1, 1),
        )
        assert p == "inside_legend_90pct_top_right"

    def test_justification_with_blank_title_and_text(self):
        p = (
            self.g
            + aes(shape="factor(cyl)")
            + guides(color=guide_colorbar(position="top"))
            + theme(
                legend_justification_top="left",
                legend_justification_right="top",
                legend_text=element_blank(),
                legend_title=element_blank(),
            )
        )
        assert p == "justification_with_blank_title_and_text"



"""
Theme elements used to decorate the graph.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable, Literal

    from plotnine import theme

    from .element_base import element_base


@dataclass
class Margin:
    element: element_base
    t: float = 0
    b: float = 0
    l: float = 0
    r: float = 0
    units: Literal["pt", "in", "lines", "fig"] = "pt"

    def __post_init__(self):
        self.theme: theme
        self.themeable_name: str

        if self.units in ("pts", "points", "px", "pixels"):
            self.units = "pt"
        elif self.units in ("in", "inch", "inches"):
            self.units = "in"
        elif self.units in ("line", "lines"):
            self.units = "lines"

    def __eq__(self, other: object) -> bool:
        def _size(m: Margin):
            return m.element.properties.get("size")

        return other is self or (
            isinstance(other, type(self))
            and other.t == self.t
            and other.b == self.b
            and other.l == self.l
            and other.r == self.r
            and other.units == self.units
            and _size(other) == _size(self)
        )

    def get_as(
        self,
        loc: Literal["t", "b", "l", "r"],
        units: Literal["pt", "in", "lines", "fig"] = "pt",
    ) -> float:
        """
        Return key in given units
        """
        dpi = 72
        size: float = self.theme.getp((self.themeable_name, "size"), 11)
        from_units = self.units
        to_units = units
        W: float
        H: float
        W, H = self.theme.getp("figure_size")  # inches
        L = (W * dpi) if loc in "tb" else (H * dpi)  # pts

        functions: dict[str, Callable[[float], float]] = {
            "fig-in": lambda x: x * L / dpi,
            "fig-lines": lambda x: x * L / size,
            "fig-pt": lambda x: x * L,
            "in-fig": lambda x: x * dpi / L,
            "in-lines": lambda x: x * dpi / size,
            "in-pt": lambda x: x * dpi,
            "lines-fig": lambda x: x * size / L,
            "lines-in": lambda x: x * size / dpi,
            "lines-pt": lambda x: x * size,
            "pt-fig": lambda x: x / L,
            "pt-in": lambda x: x / dpi,
            "pt-lines": lambda x: x / size,
        }

        value: float = getattr(self, loc)
        if from_units != to_units:
            conversion = f"{self.units}-{units}"
            try:
                value = functions[conversion](value)
            except ZeroDivisionError:
                value = 0
        return value



from .elements import element_blank, element_line, element_rect, element_text
from .theme import theme
from .theme_gray import theme_gray


class theme_xkcd(theme_gray):
    """
    xkcd theme

    Parameters
    ----------
    base_size : int
        Base font size. All text sizes are a scaled versions of
        the base font size.
    scale : float
        The amplitude of the wiggle perpendicular to the line (in pixels)
    length : float
        The length of the wiggle along the line (in pixels).
    randomness : float
        The factor by which the length is randomly scaled. Default is 2.
    stroke_size : float
        Size of the stroke to apply to the lines and text paths.
    stroke_color : str | tuple
        Color of the strokes. Use `"none"` for no color.
    """

    def __init__(
        self,
        base_size=12,
        scale=1,
        length=100,
        randomness=2,
        stroke_size=3,
        stroke_color="white",
    ):
        from matplotlib import patheffects

        super().__init__(base_size)
        sketch_params = (scale, length, randomness)
        path_effects = [
            patheffects.withStroke(
                linewidth=stroke_size, foreground=stroke_color
            )
        ]
        self += theme(
            text=element_text(family=["xkcd", "Humor Sans", "Comic Sans MS"]),
            axis_ticks=element_line(color="black", size=1),
            axis_ticks_minor=element_blank(),
            axis_ticks_direction="in",
            axis_ticks_length_major=6,
            legend_background=element_rect(color="black"),
            legend_box_margin=2,
            legend_margin=5,
            legend_key=element_rect(fill="none"),
            panel_border=element_rect(color="black", size=1),
            panel_grid=element_blank(),
            panel_background=element_rect(fill="white"),
            strip_background=element_rect(color="black", fill="white"),
            strip_background_x=element_rect(width=2 / 3),
            strip_background_y=element_rect(height=2 / 3),
            strip_align=-0.5,
        )

        self._rcParams.update(
            {
                "axes.unicode_minus": False,
                "path.sketch": sketch_params,
                "path.effects": path_effects,
            }
        )



from __future__ import annotations

import typing
from abc import ABC
from copy import deepcopy
from itertools import chain, repeat

from .._utils import (
    copy_keys,
    data_mapping_as_kwargs,
    is_list_like,
    remove_missing,
)
from .._utils.registry import Register, Registry
from ..exceptions import PlotnineError
from ..layer import layer
from ..mapping.aes import is_valid_aesthetic, rename_aesthetics
from ..mapping.evaluation import evaluate
from ..positions.position import position
from ..stats.stat import stat

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd
    from matplotlib.axes import Axes
    from matplotlib.offsetbox import DrawingArea

    from plotnine import aes, ggplot
    from plotnine.coords.coord import coord
    from plotnine.facets.layout import Layout
    from plotnine.iapi import panel_view
    from plotnine.mapping import Environment
    from plotnine.typing import DataLike, TupleInt2


class geom(ABC, metaclass=Register):
    """Base class of all Geoms"""

    DEFAULT_AES: dict[str, Any] = {}
    """Default aesthetics for the geom"""

    REQUIRED_AES: set[str] = set()
    """Required aesthetics for the geom"""

    NON_MISSING_AES: set[str] = set()
    """Required aesthetics for the geom"""

    DEFAULT_PARAMS: dict[str, Any] = {}
    """Required parameters for the geom"""

    data: DataLike
    """Geom/layer specific dataframe"""

    mapping: aes
    """Mappings i.e. `aes(x="col1", fill="col2")`{.py}"""

    aes_params: dict[str, Any] = {}  # setting of aesthetic
    params: dict[str, Any]  # parameter settings

    # Plot namespace, it gets its value when the plot is being
    # built.
    environment: Environment

    # The geom responsible for the legend if draw_legend is
    # not implemented
    legend_geom: str = "point"

    # Documentation for the aesthetics. It is added under the
    # documentation for mapping parameter. Use {aesthetics}
    # placeholder to insert a table for all the aesthetics and
    # their default values.
    _aesthetics_doc: str = "{aesthetics_table}"

    def __init__(
        self,
        mapping: aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        kwargs = rename_aesthetics(kwargs)
        kwargs = data_mapping_as_kwargs((data, mapping), kwargs)
        self._kwargs = kwargs  # Will be used to create stat & layer

        # separate aesthetics and parameters
        self.aes_params = copy_keys(kwargs, {}, self.aesthetics())
        self.params = copy_keys(kwargs, deepcopy(self.DEFAULT_PARAMS))
        self.mapping = kwargs["mapping"]
        self.data = kwargs["data"]
        self._stat = stat.from_geom(self)
        self._position = position.from_geom(self)
        self._verify_arguments(kwargs)  # geom, stat, layer

    @staticmethod
    def from_stat(stat: stat) -> geom:
        """
        Return an instantiated geom object

        geoms should not override this method.

        Parameters
        ----------
        stat :
            `stat`

        Returns
        -------
        :
            A geom object

        Raises
        ------
        PlotnineError
            If unable to create a `geom`.
        """
        name = stat.params["geom"]

        if isinstance(name, geom):
            return name

        if isinstance(name, type) and issubclass(name, geom):
            klass = name
        elif isinstance(name, str):
            if not name.startswith("geom_"):
                name = f"geom_{name}"
            klass = Registry[name]
        else:
            raise PlotnineError(f"Unknown geom of type {type(name)}")

        return klass(stat=stat, **stat._kwargs)

    @classmethod
    def aesthetics(cls: type[geom]) -> set[str]:
        """
        Return all the aesthetics for this geom

        geoms should not override this method.
        """
        main = cls.DEFAULT_AES.keys() | cls.REQUIRED_AES
        other = {"group"}
        # Need to recognize both spellings
        if "color" in main:
            other.add("colour")
        if "outlier_color" in main:
            other.add("outlier_colour")
        return main | other

    def __deepcopy__(self, memo: dict[Any, Any]) -> geom:
        """
        Deep copy without copying the self.data dataframe

        geoms should not override this method.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        # don't make a deepcopy of data, or environment
        shallow = {"data", "_kwargs", "environment"}
        for key, item in old.items():
            if key in shallow:
                new[key] = item
                memo[id(new[key])] = new[key]
            else:
                new[key] = deepcopy(item, memo)

        return result

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Modify the data before drawing takes place

        This function is called *before* position adjustments are done.
        It is used by geoms to create the final aesthetics used for
        drawing. The base class method does nothing, geoms can override
        this method for two reasons:

        1. The `stat` does not create all the aesthetics (usually
           position aesthetics) required for drawing the `geom`,
           but those aesthetics can be computed from the available
           data. For example [](`~plotnine.geoms.geom_boxplot`)
           and [](`~plotnine.geoms.geom_violin`).

        2. The `geom` inherits from another `geom` (superclass) which
           does the drawing and the superclass requires certain aesthetics
           to be present in the data. For example
           [](`~plotnine.geoms.geom_tile`) and
           [](`~plotnine.geoms.geom_area`).

        Parameters
        ----------
        data :
            Data used for drawing the geom.

        Returns
        -------
        :
            Data used for drawing the geom.
        """
        return data

    def use_defaults(
        self, data: pd.DataFrame, aes_modifiers: dict[str, Any]
    ) -> pd.DataFrame:
        """
        Combine data with defaults and set aesthetics from parameters

        geoms should not override this method.

        Parameters
        ----------
        data :
            Data used for drawing the geom.
        aes_modifiers :
            Aesthetics

        Returns
        -------
        :
            Data used for drawing the geom.
        """
        missing_aes = (
            self.DEFAULT_AES.keys()
            - self.aes_params.keys()
            - set(data.columns.to_list())
        )

        # Not in data and not set, use default
        for ae in missing_aes:
            data[ae] = self.DEFAULT_AES[ae]

        # Evaluate/Modify the mapped aesthetics
        evaled = evaluate(aes_modifiers, data, self.environment)
        for ae in evaled.columns.intersection(data.columns):
            data[ae] = evaled[ae]

        num_panels = len(data["PANEL"].unique()) if "PANEL" in data else 1

        # Aesthetics set as parameters to the geom/stat
        for ae, value in self.aes_params.items():
            try:
                data[ae] = value
            except ValueError as e:
                # sniff out the special cases, like custom
                # tupled linetypes, shapes and colors
                if is_valid_aesthetic(value, ae):
                    data[ae] = [value] * len(data)
                elif num_panels > 1 and is_list_like(value):
                    data[ae] = list(chain(*repeat(value, num_panels)))
                else:
                    msg = (
                        f"'{value}' does not look like a "
                        f"valid value for `{ae}`"
                    )
                    raise PlotnineError(msg) from e

        return data

    def draw_layer(
        self, data: pd.DataFrame, layout: Layout, coord: coord, **params: Any
    ):
        """
        Draw layer across all panels

        geoms should not override this method.

        Parameters
        ----------
        data :
            DataFrame specific for this layer
        layout :
            Layout object created when the plot is getting
            built
        coord :
            Type of coordinate axes
        params :
            Combined *geom* and *stat* parameters. Also
            includes the stacking order of the layer in
            the plot (*zorder*)
        """
        for pid, pdata in data.groupby("PANEL", observed=True):
            if len(pdata) == 0:
                continue
            ploc = pdata["PANEL"].iloc[0] - 1
            panel_params = layout.panel_params[ploc]
            ax = layout.axs[ploc]
            self.draw_panel(pdata, panel_params, coord, ax, **params)

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups

        For efficiency, geoms that do not need to partition
        different groups before plotting should override this
        method and avoid the groupby.

        Parameters
        ----------
        data :
            Data to be plotted by this geom. This is the
            dataframe created in the plot_build pipeline.
        panel_params :
            The scale information as may be required by the
            axes. At this point, that information is about
            ranges, ticks and labels. Attributes are of interest
            to the geom are:

            ```python
            "panel_params.x.range"  # tuple
            "panel_params.y.range"  # tuple
            ```
        coord :
            Coordinate (e.g. coord_cartesian) system of the geom.
        ax :
            Axes on which to plot.
        params :
            Combined parameters for the geom and stat. Also
            includes the `zorder`.
        """
        for _, gdata in data.groupby("group"):
            gdata.reset_index(inplace=True, drop=True)
            self.draw_group(gdata, panel_params, coord, ax, **params)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot data belonging to a group.

        Parameters
        ----------
        data :
            Data to be plotted by this geom. This is the
            dataframe created in the plot_build pipeline.
        panel_params :
            The scale information as may be required by the
            axes. At this point, that information is about
            ranges, ticks and labels. Keys of interest to
            the geom are:

            ```python
            "x_range"  # tuple
            "y_range"  # tuple
            ```
        coord : coord
            Coordinate (e.g. coord_cartesian) system of the geom.
        ax : axes
            Axes on which to plot.
        params : dict
            Combined parameters for the geom and stat. Also
            includes the `zorder`.
        """
        msg = "The geom should implement this method."
        raise NotImplementedError(msg)

    @staticmethod
    def draw_unit(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot data belonging to a unit.

        A matplotlib plot function may require that an aethestic
        have a single unique value. e.g. `linestyle="dashed"`{.py}
        and not `linestyle=["dashed", "dotted", ...]`{.py}.
        A single call to such a function can only plot lines with
        the same linestyle. However, if the plot we want has more
        than one line with different linestyles, we need to group
        the lines with the same linestyle and plot them as one
        unit. In this case, draw_group calls this function to do
        the plotting. For an example see
        [](`~plotnine.geoms.geom_point`).

        Parameters
        ----------
        data :
            Data to be plotted by this geom. This is the
            dataframe created in the plot_build pipeline.
        panel_params :
            The scale information as may be required by the
            axes. At this point, that information is about
            ranges, ticks and labels. Keys of interest to
            the geom are:

            ```python
            "x_range"  # tuple
            "y_range"  # tuple
            ```

            In rare cases a geom may need access to the x or y scales.
            Those are available at:

            ```python
            "scales"   # SimpleNamespace
            ```
        coord :
            Coordinate (e.g. coord_cartesian) system of the
            geom.
        ax :
            Axes on which to plot.
        params :
            Combined parameters for the geom and stat. Also
            includes the `zorder`.
        """
        msg = "The geom should implement this method."
        raise NotImplementedError(msg)

    def __radd__(self, plot: ggplot) -> ggplot:
        """
        Add layer representing geom object on the right

        Parameters
        ----------
        plot :
            ggplot object

        Returns
        -------
        :
            ggplot object with added layer.
        """
        plot += self.to_layer()  # Add layer
        return plot

    def to_layer(self) -> layer:
        """
        Make a layer that represents this geom

        Returns
        -------
        :
            Layer
        """
        return layer.from_geom(self)

    def _verify_arguments(self, kwargs: dict[str, Any]):
        """
        Verify arguments passed to the geom
        """
        geom_stat_args = kwargs.keys() | self._stat._kwargs.keys()
        unknown = (
            geom_stat_args
            - self.aesthetics()
            - self.DEFAULT_PARAMS.keys()  # geom aesthetics
            - self._stat.aesthetics()  # geom parameters
            - self._stat.DEFAULT_PARAMS.keys()  # stat aesthetics
            - {  # stat parameters
                "data",
                "mapping",
                "show_legend",  # layer parameters
                "inherit_aes",
                "raster",
            }
        )  # layer parameters
        if unknown:
            msg = (
                "Parameters {}, are not understood by "
                "either the geom, stat or layer."
            )
            raise PlotnineError(msg.format(unknown))

    def handle_na(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Remove rows with NaN values

        geoms that infer extra information from missing values
        should override this method. For example
        [](`~plotnine.geoms.geom_path`).

        Parameters
        ----------
        data :
            Data

        Returns
        -------
        :
            Data without the NaNs.

        Notes
        -----
        Shows a warning if the any rows are removed and the
        `na_rm` parameter is False. It only takes into account
        the columns of the required aesthetics.
        """
        return remove_missing(
            data,
            self.params["na_rm"],
            list(self.REQUIRED_AES | self.NON_MISSING_AES),
            self.__class__.__name__,
        )

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: layer
    ) -> DrawingArea:
        """
        Draw a rectangle in the box

        Parameters
        ----------
        data :
            A row of the data plotted to this layer
        da :
            Canvas on which to draw
        lyr :
            Layer that the geom belongs to.

        Returns
        -------
        :
            The DrawingArea after a layer has been drawn onto it.
        """
        msg = "The geom should implement this method."
        raise NotImplementedError(msg)

    @staticmethod
    def legend_key_size(
        data: pd.Series[Any], min_size: TupleInt2, lyr: layer
    ) -> TupleInt2:
        """
        Calculate the size of key that would fit the layer contents

        Parameters
        ----------
        data :
            A row of the data plotted to this layer
        min_size :
            Initial size which should be expanded to fit the contents.
        lyr :
            Layer
        """
        return min_size



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional

    from matplotlib.collections import LineCollection
    from matplotlib.patches import Rectangle
    from matplotlib.text import Text

    from plotnine._mpl.offsetbox import ColoredDrawingArea
    from plotnine._mpl.patches import StripTextPatch
    from plotnine._mpl.text import StripText
    from plotnine.iapi import legend_artists


@dataclass
class ThemeTargets:
    """
    Artists that will be themed

    This includes only artist that cannot be accessed easily from
    the figure or the axes.
    """

    axis_title_x: Optional[Text] = None
    axis_title_y: Optional[Text] = None
    legend_frame: Optional[Rectangle] = None
    legend_key: list[ColoredDrawingArea] = field(default_factory=list)
    legends: Optional[legend_artists] = None
    legend_text_colorbar: list[Text] = field(default_factory=list)
    legend_text_legend: list[Text] = field(default_factory=list)
    legend_ticks: Optional[LineCollection] = None
    legend_title: Optional[Text] = None
    panel_border: list[Rectangle] = field(default_factory=list)
    plot_caption: Optional[Text] = None
    plot_subtitle: Optional[Text] = None
    plot_title: Optional[Text] = None
    strip_background_x: list[StripTextPatch] = field(default_factory=list)
    strip_background_y: list[StripTextPatch] = field(default_factory=list)
    strip_text_x: list[StripText] = field(default_factory=list)
    strip_text_y: list[StripText] = field(default_factory=list)



from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from .._utils import copy_missing_columns, resolution
from ..doctools import document
from .geom import geom
from .geom_path import geom_path
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    from typing import Any

    from matplotlib.axes import Axes

    from plotnine.coords.coord import coord
    from plotnine.iapi import panel_view


@document
class geom_errorbarh(geom):
    """
    Horizontal interval represented as an errorbar

    {usage}

    Parameters
    ----------
    {common_parameters}
    height : float, default=0.5
        Bar height as a fraction of the resolution of the data.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "black",
        "linetype": "solid",
        "size": 0.5,
    }
    REQUIRED_AES = {"y", "xmin", "xmax"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "height": 0.5,
    }
    draw_legend = staticmethod(geom_path.draw_legend)

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if "height" not in data:
            if self.params["height"]:
                data["height"] = self.params["height"]
            else:
                data["height"] = resolution(data["y"], False) * 0.9

        data["ymin"] = data["y"] - data["height"] / 2
        data["ymax"] = data["y"] + data["height"] / 2
        del data["height"]
        return data

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: coord,
        ax: Axes,
        **params: Any,
    ):
        f = np.hstack
        # create (two vertical bars) + horizontal bar
        bars = pd.DataFrame(
            {
                "y": f([data["ymin"], data["ymin"], data["y"]]),
                "yend": f([data["ymax"], data["ymax"], data["y"]]),
                "x": f([data["xmin"], data["xmax"], data["xmin"]]),
                "xend": f([data["xmin"], data["xmax"], data["xmax"]]),
            }
        )

        copy_missing_columns(bars, data)
        geom_segment.draw_group(bars, panel_params, coord, ax, **params)



from __future__ import annotations

from dataclasses import dataclass, fields
from typing import TYPE_CHECKING

from .iapi import labels_view
from .mapping.aes import rename_aesthetics

if TYPE_CHECKING:
    import plotnine as p9

__all__ = ("xlab", "ylab", "labs", "ggtitle")


# TODO: Make keyword-only when python>=3.10
# i.e. @dataclass(kw_only=True)
@dataclass
class labs:
    """
    Add labels for any aesthetics with a scale or title, subtitle & caption
    """

    # Names of Scaled Aesthetics
    x: str | None = None
    """
    Name of the x-axis.
    """

    y: str | None = None
    """
    Name of the y-axis.
    """

    alpha: str | None = None
    """
    Name of the alpha legend.
    """

    color: str | None = None
    """
    Name of the color legend or colorbar.
    """

    colour: str | None = None
    """
    Name of the colour legend or colourbar.

    This is an alias of the `color` parameter. Only use one of
    the spellings.
    """

    fill: str | None = None
    """
    Name of the fill legend/colourbar.
    """

    linetype: str | None = None
    """
    Name of the linetype legend.
    """

    shape: str | None = None
    """
    Name of the shape legend.
    """

    size: str | None = None
    """
    Name of the size legend.
    """

    stroke: str | None = None
    """
    Name of the stroke legend.
    """

    # Other texts
    title: str | None = None
    """
    The title of the plot.
    """

    subtitle: str | None = None
    """
    The subtitle of the plot.
    """

    caption: str | None = None
    """
    The caption at the bottom of the plot.
    """

    def __post_init__(self):
        kwargs: dict[str, str] = {
            f.name: value
            for f in fields(self)
            if (value := getattr(self, f.name)) is not None
        }
        self.labels = labels_view(**rename_aesthetics(kwargs))

    def __radd__(self, plot: p9.ggplot) -> p9.ggplot:
        """
        Add labels to ggplot object
        """
        plot.labels.update(self.labels)
        return plot


class xlab(labs):
    """
    Label/name for the x aesthetic

    Parameters
    ----------
    name :
        x aesthetic label (x-axis)
    """

    def __init__(self, name: str, /):
        self.x = name
        self.labels = labels_view(x=name)


class ylab(labs):
    """
    Label/name for the y aesthetic

    Parameters
    ----------
    name :
        y aesthetic label i.e. y-axis label
    """

    def __init__(self, name: str, /):
        self.y = name
        self.labels = labels_view(y=name)


class ggtitle(labs):
    """
    Create plot title

    Parameters
    ----------
    title :
        Plot title
    """

    def __init__(self, title: str, /):
        self.title = title
        self.labels = labels_view(title=title)



import warnings

from plotnine import (
    aes,
    geom_point,
    ggplot,
)
from plotnine.data import mtcars


def test_no_after_scale_warning():
    p = ggplot(mtcars, aes("wt", "mpg")) + geom_point()

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        p.draw_test()  # type: ignore



"""
Aesthetic Mappings
"""

from ._env import Environment  # noqa: F401
from .aes import aes
from .evaluation import after_scale, after_stat, stage

__all__ = ("aes", "after_stat", "after_scale", "stage")



from .element_blank import element_blank
from .element_line import element_line
from .element_rect import element_rect
from .element_text import element_text

__all__ = (
    "element_blank",
    "element_line",
    "element_rect",
    "element_text",
)
